from typing import Any, Mapping, Union

import sqlalchemy as sa
from lark import Lark, LarkError, Transformer, Tree

from . import FieldSpecItem

__all__ = (
    "FilterableSQLQuery",
    "QueryFilterParser",
)


FilterableSQLQuery = Union[sa.sql.Select, sa.sql.Update, sa.sql.Delete]

_grammar = r"""
    ?start: expr
    value: string
           | number
           | array
           | ATOM -> atom
    ATOM       : "null" | "true" | "false"
    COMBINE_OP : "&" | "|"
    UNARY_OP   : "!"
    BINARY_OP  : "==" | "!="
               | ">" | ">="
               | "<" | "<="
               | "contains" | "in"
               | "isnot" | "is"
               | "like" | "ilike"
    expr: UNARY_OP expr         -> unary_expr
        | CNAME BINARY_OP value -> binary_expr
        | expr COMBINE_OP expr  -> combine_expr
        | "(" expr ")"          -> paren_expr
    array  : "[" [value ("," value)*] "]"
    string : ESCAPED_STRING
    number : SIGNED_NUMBER
    %import common.CNAME
    %import common.ESCAPED_STRING
    %import common.SIGNED_NUMBER
    %import common.WS
    %ignore WS
"""
_parser = Lark(
    _grammar,
    parser="lalr",
    maybe_placeholders=False,
)


class QueryFilterTransformer(Transformer):
    def __init__(self, sa_table: sa.Table, fieldspec: Mapping[str, FieldSpecItem] = None) -> None:
        super().__init__()
        self._sa_table = sa_table
        self._fieldspec = fieldspec

    def string(self, s):
        (s,) = s
        # SQL-side escaping is handled by SQLAlchemy
        return s[1:-1].replace('\\"', '"')

    def number(self, n):
        (n,) = n
        if "." in n:
            return float(n)
        return int(n)

    array = list

    def atom(self, a):
        (a,) = a
        if a.value == "null":
            return sa.null()
        elif a.value == "true":
            return sa.true()
        elif a.value == "false":
            return sa.false()

    def _get_col(self, col_name: str) -> sa.Column:
        try:
            if self._fieldspec:
                col = self._sa_table.c[self._fieldspec[col_name][0]]
            else:
                col = self._sa_table.c[col_name]
            return col
        except KeyError:
            raise ValueError("Unknown/unsupported field name", col_name)

    def _transform_val_leaf(self, col_name: str, value: Any) -> Any:
        if self._fieldspec:
            try:
                func = self._fieldspec[col_name][1]
            except KeyError:
                raise ValueError("Unknown/unsupported field name", col_name)
            return func(value) if func is not None else value
        else:
            return value

    def _transform_val(self, col_name: str, value: Any) -> Any:
        if isinstance(value, Tree):
            val = self._transform_val(col_name, value.children[0])
        elif isinstance(value, list):
            val = [self._transform_val(col_name, v) for v in value]
        else:
            val = self._transform_val_leaf(col_name, value)
        return val

    def binary_expr(self, *args):
        children = args[0]
        col = self._get_col(children[0].value)
        op = children[1].value
        val = self._transform_val(children[0].value, children[2])
        if op == "==":
            return col == val
        elif op == "!=":
            return col != val
        elif op == ">":
            return col > val
        elif op == ">=":
            return col >= val
        elif op == "<":
            return col < val
        elif op == "<=":
            return col <= val
        elif op == "contains":
            return col.contains(val)
        elif op == "in":
            return col.in_(val)
        elif op == "isnot":
            return col.isnot(val)
        elif op == "is":
            return col.is_(val)
        elif op == "like":
            return col.like(val)
        elif op == "ilike":
            return col.ilike(val)
        return args

    def unary_expr(self, *args):
        children = args[0]
        op = children[0].value
        expr = children[1]
        if op in ("not", "!"):
            return sa.not_(expr)
        return args

    def combine_expr(self, *args):
        children = args[0]
        op = children[1].value
        expr1 = children[0]
        expr2 = children[2]
        if op == "&":
            return sa.and_(expr1, expr2)
        elif op == "|":
            return sa.or_(expr1, expr2)
        return args

    def paren_expr(self, *args):
        children = args[0]
        return children[0]


class QueryFilterParser:
    def __init__(self, fieldspec: Mapping[str, FieldSpecItem] = None) -> None:
        self._fieldspec = fieldspec
        self._parser = _parser

    def append_filter(
        self,
        sa_query: FilterableSQLQuery,
        filter_expr: str,
    ) -> FilterableSQLQuery:
        """
        Parse the given filter expression and build the where clause based on the first target table from
        the given SQLAlchemy query object.
        """
        if isinstance(sa_query, sa.sql.Select):
            table = sa_query.froms[0]
        elif isinstance(sa_query, sa.sql.Delete):
            table = sa_query.table
        elif isinstance(sa_query, sa.sql.Update):
            table = sa_query.table
        else:
            raise ValueError("Unsupported SQLAlchemy query object type")
        try:
            ast = self._parser.parse(filter_expr)
            where_clause = QueryFilterTransformer(table, self._fieldspec).transform(ast)
        except LarkError as e:
            raise ValueError(f"Query filter parsing error: {e}")
        return sa_query.where(where_clause)



"""add_modified_at_to_users_and_kernels

Revision ID: e35332f8d23d
Revises: da24ff520049
Create Date: 2020-07-01 14:02:11.022032

"""
import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql
from sqlalchemy.sql.expression import bindparam

from ai.backend.manager.models.base import IDColumn, convention

# revision identifiers, used by Alembic.
revision = "e35332f8d23d"
down_revision = "da24ff520049"
branch_labels = None
depends_on = None


def upgrade():
    metadata = sa.MetaData(naming_convention=convention)
    # partial table to be preserved and referred
    users = sa.Table(
        "users",
        metadata,
        IDColumn("uuid"),
        sa.Column("created_at", sa.DateTime(timezone=True), server_default=sa.func.now()),
        sa.Column(
            "modified_at",
            sa.DateTime(timezone=True),
            server_default=sa.func.now(),
            onupdate=sa.func.current_timestamp(),
        ),
    )
    keypairs = sa.Table(
        "keypairs",
        metadata,
        sa.Column("access_key", sa.String(length=20), primary_key=True),
        sa.Column("created_at", sa.DateTime(timezone=True), server_default=sa.func.now()),
        sa.Column(
            "modified_at",
            sa.DateTime(timezone=True),
            server_default=sa.func.now(),
            onupdate=sa.func.current_timestamp(),
        ),
    )

    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "keypairs",
        sa.Column(
            "modified_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
    )
    op.add_column(
        "users",
        sa.Column(
            "modified_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
    )
    # ### end Alembic commands ###

    conn = op.get_bind()

    # Set user's modified_at with the value of created_at.
    query = sa.select([users.c.uuid, users.c.created_at]).select_from(users)
    updates = []
    for row in conn.execute(query).fetchall():
        updates.append({"b_uuid": row["uuid"], "modified_at": row["created_at"]})
    if updates:
        query = (
            sa.update(users)
            .values(modified_at=bindparam("modified_at"))
            .where(users.c.uuid == bindparam("b_uuid"))
        )
        conn.execute(query, updates)

    # Set keypairs's modified_at with the value of created_at.
    query = sa.select([keypairs.c.access_key, keypairs.c.created_at]).select_from(keypairs)
    updates = []
    for row in conn.execute(query).fetchall():
        updates.append({"b_access_key": row["access_key"], "modified_at": row["created_at"]})
    if updates:
        query = (
            sa.update(keypairs)
            .values(modified_at=bindparam("modified_at"))
            .where(keypairs.c.access_key == bindparam("b_access_key"))
        )
        conn.execute(query, updates)


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("users", "modified_at")
    op.drop_column("keypairs", "modified_at")
    # ### end Alembic commands ###



from __future__ import annotations

import asyncio
import json
import logging
from abc import ABCMeta, abstractmethod
from contextvars import ContextVar
from typing import Any, AsyncIterator, Dict, Mapping, Optional, cast

import aiohttp
import aiotools
import sqlalchemy as sa
import yarl

from ai.backend.common.bgtask import ProgressReporter
from ai.backend.common.docker import MAX_KERNELSPEC, MIN_KERNELSPEC, ImageRef, arch_name_aliases
from ai.backend.common.docker import login as registry_login
from ai.backend.common.logging import BraceStyleAdapter
from ai.backend.manager.models.image import ImageRow, ImageType
from ai.backend.manager.models.utils import ExtendedAsyncSAEngine

log = BraceStyleAdapter(logging.getLogger(__name__))


class BaseContainerRegistry(metaclass=ABCMeta):

    db: ExtendedAsyncSAEngine
    registry_name: str
    registry_info: Mapping[str, Any]
    registry_url: yarl.URL
    max_concurrency_per_registry: int
    base_hdrs: Dict[str, str]
    credentials: Dict[str, str]
    ssl_verify: bool

    sema: ContextVar[asyncio.Semaphore]
    reporter: ContextVar[Optional[ProgressReporter]]
    all_updates: ContextVar[Dict[ImageRef, Dict[str, Any]]]

    def __init__(
        self,
        db: ExtendedAsyncSAEngine,
        registry_name: str,
        registry_info: Mapping[str, Any],
        *,
        max_concurrency_per_registry: int = 4,
        ssl_verify: bool = True,
    ) -> None:
        self.db = db
        self.registry_name = registry_name
        self.registry_info = registry_info
        self.registry_url = registry_info[""]
        self.max_concurrency_per_registry = max_concurrency_per_registry
        self.base_hdrs = {
            "Accept": "application/vnd.docker.distribution.manifest.v2+json",
        }
        self.credentials = {}
        self.ssl_verify = ssl_verify
        self.sema = ContextVar("sema")
        self.reporter = ContextVar("reporter", default=None)
        self.all_updates = ContextVar("all_updates")

    async def rescan_single_registry(
        self,
        reporter: ProgressReporter = None,
    ) -> None:
        self.all_updates.set({})
        self.sema.set(asyncio.Semaphore(self.max_concurrency_per_registry))
        self.reporter.set(reporter)
        username = self.registry_info["username"]
        if username is not None:
            self.credentials["username"] = username
        password = self.registry_info["password"]
        if password is not None:
            self.credentials["password"] = password
        non_kernel_words = (
            "common-",
            "commons-",
            "base-",
            "krunner",
            "builder",
            "backendai",
            "geofront",
        )
        ssl_ctx = None  # default
        if not self.registry_info["ssl-verify"]:
            ssl_ctx = False
        connector = aiohttp.TCPConnector(ssl=ssl_ctx)
        async with aiohttp.ClientSession(connector=connector) as sess:
            async with aiotools.TaskGroup() as tg:
                async for image in self.fetch_repositories(sess):
                    if not any((w in image) for w in non_kernel_words):  # skip non-kernel images
                        tg.create_task(self._scan_image(sess, image))

        all_updates = self.all_updates.get()
        if not all_updates:
            log.info("No images found in registry {0}", self.registry_url)
        else:
            image_identifiers = [(k.canonical, k.architecture) for k in all_updates.keys()]
            async with self.db.begin_session() as session:
                existing_images = await session.scalars(
                    sa.select(ImageRow).where(
                        sa.func.ROW(ImageRow.name, ImageRow.architecture).in_(image_identifiers),
                    ),
                )

                for image_row in existing_images:
                    key = image_row.image_ref
                    values = all_updates.get(key)
                    if values is None:
                        continue
                    all_updates.pop(key)
                    image_row.config_digest = values["config_digest"]
                    image_row.size_bytes = values["size_bytes"]
                    image_row.accelerators = values.get("accels")
                    image_row.labels = values["labels"]
                    image_row.resources = values["resources"]

                session.add_all(
                    [
                        ImageRow(
                            name=k.canonical,
                            registry=k.registry,
                            image=k.name,
                            tag=k.tag,
                            architecture=k.architecture,
                            config_digest=v["config_digest"],
                            size_bytes=v["size_bytes"],
                            type=ImageType.COMPUTE,
                            accelerators=v.get("accels"),
                            labels=v["labels"],
                            resources=v["resources"],
                        )
                        for k, v in all_updates.items()
                    ]
                )

    async def _scan_image(
        self,
        sess: aiohttp.ClientSession,
        image: str,
    ) -> None:
        rqst_args = await registry_login(
            sess,
            self.registry_url,
            self.credentials,
            f"repository:{image}:pull",
        )
        rqst_args["headers"].update(**self.base_hdrs)
        tags = []
        tag_list_url: Optional[yarl.URL]
        tag_list_url = (self.registry_url / f"v2/{image}/tags/list").with_query(
            {"n": "10"},
        )
        while tag_list_url is not None:
            async with sess.get(tag_list_url, **rqst_args) as resp:
                data = json.loads(await resp.read())
                if "tags" in data:
                    # sometimes there are dangling image names in the hub.
                    tags.extend(data["tags"])
                tag_list_url = None
                next_page_link = resp.links.get("next")
                if next_page_link:
                    next_page_url = cast(yarl.URL, next_page_link["url"])
                    tag_list_url = self.registry_url.with_path(next_page_url.path).with_query(
                        next_page_url.query
                    )
        if (reporter := self.reporter.get()) is not None:
            reporter.total_progress += len(tags)
        async with aiotools.TaskGroup() as tg:
            for tag in tags:
                tg.create_task(self._scan_tag(sess, rqst_args, image, tag))

    async def _scan_tag(
        self,
        sess: aiohttp.ClientSession,
        rqst_args,
        image: str,
        tag: str,
    ) -> None:
        skip_reason = None

        async def _load_manifest(_tag: str):
            async with sess.get(
                self.registry_url / f"v2/{image}/manifests/{_tag}", **rqst_args
            ) as resp:
                if resp.status == 404:
                    # ignore missing tags
                    # (may occur after deleting an image from the docker hub)
                    return {}
                resp.raise_for_status()
                data = await resp.json()

                if data["mediaType"] == "application/vnd.docker.distribution.manifest.list.v2+json":
                    # recursively call _load_manifests with detected arch and corresponding image digest
                    ret = {}
                    for m in data["manifests"]:
                        ret.update(
                            await _load_manifest(
                                m["digest"],
                            ),
                        )
                    if (reporter := self.reporter.get()) is not None:
                        reporter.total_progress += len(ret) - 1
                    return ret

                config_digest = data["config"]["digest"]
                size_bytes = sum(layer["size"] for layer in data["layers"]) + data["config"]["size"]
                async with sess.get(
                    self.registry_url / f"v2/{image}/blobs/{config_digest}", **rqst_args
                ) as resp:
                    resp.raise_for_status()
                    data = json.loads(await resp.read())
                    architecture = arch_name_aliases.get(data["architecture"], data["architecture"])
                    labels = {}
                    if "container_config" in data:
                        raw_labels = data["container_config"].get("Labels")
                        if raw_labels:
                            labels.update(raw_labels)
                        else:
                            log.warn("label not found on image {}:{}/{}", image, _tag, architecture)
                    else:
                        raw_labels = data["config"].get("Labels")
                        if raw_labels:
                            labels.update(raw_labels)
                        else:
                            log.warn("label not found on image {}:{}/{}", image, _tag, architecture)
                    return {
                        architecture: {
                            "size": size_bytes,
                            "labels": labels,
                            "digest": config_digest,
                        },
                    }

        async with self.sema.get():
            manifests = await _load_manifest(tag)

        if len(manifests.keys()) == 0:
            log.warning("Skipped image - {}:{} (missing/deleted)", image, tag)
            progress_msg = f"Skipped {image}:{tag} (missing/deleted)"
            if (reporter := self.reporter.get()) is not None:
                await reporter.update(1, message=progress_msg)

        idx = 0
        for architecture, manifest in manifests.items():
            idx += 1
            if manifest is None:
                skip_reason = "missing/deleted"
                continue

            try:
                size_bytes = manifest["size"]
                labels = manifest["labels"]
                config_digest = manifest["digest"]
                if "ai.backend.kernelspec" not in labels:
                    # Skip non-Backend.AI kernel images
                    skip_reason = architecture + ": missing kernelspec"
                    continue
                if not (MIN_KERNELSPEC <= int(labels["ai.backend.kernelspec"]) <= MAX_KERNELSPEC):
                    # Skip unsupported kernelspec images
                    skip_reason = architecture + ": unsupported kernelspec"
                    continue

                update_key = ImageRef(
                    f"{self.registry_name}/{image}:{tag}",
                    [self.registry_name],
                    architecture,
                )
                updates = {
                    "config_digest": config_digest,
                    "size_bytes": size_bytes,
                    "labels": labels,
                }
                accels = labels.get("ai.backend.accelerators")
                if accels:
                    updates["accels"] = accels

                resources = {}
                res_prefix = "ai.backend.resource.min."
                for k, v in filter(lambda pair: pair[0].startswith(res_prefix), labels.items()):
                    res_key = k[len(res_prefix) :]
                    resources[res_key] = {"min": v}
                updates["resources"] = resources
                self.all_updates.get().update(
                    {
                        update_key: updates,
                    }
                )
            finally:
                if skip_reason:
                    log.warning(
                        "Skipped image - {}:{}/{} ({})", image, tag, architecture, skip_reason
                    )
                    progress_msg = f"Skipped {image}:{tag}/{architecture} ({skip_reason})"
                else:
                    log.info("Updated image - {0}:{1}/{2}", image, tag, architecture)
                    progress_msg = f"Updated {image}:{tag}/{architecture}"
                if (reporter := self.reporter.get()) is not None:
                    await reporter.update(1, message=progress_msg)

    @abstractmethod
    async def fetch_repositories(
        self,
        sess: aiohttp.ClientSession,
    ) -> AsyncIterator[str]:
        yield ""



from __future__ import annotations

import asyncio
import logging
import subprocess
import sys
from datetime import datetime
from functools import partial
from pathlib import Path

import click
from more_itertools import chunked
from setproctitle import setproctitle

from ai.backend.cli.types import ExitCode
from ai.backend.common import redis_helper as redis_helper
from ai.backend.common.cli import LazyGroup
from ai.backend.common.logging import BraceStyleAdapter
from ai.backend.common.validators import TimeDuration

from ..config import load as load_config
from .context import CLIContext, init_logger, redis_ctx

log = BraceStyleAdapter(logging.getLogger("ai.backend.manager.cli"))


@click.group(invoke_without_command=True, context_settings={"help_option_names": ["-h", "--help"]})
@click.option(
    "-f",
    "--config-path",
    "--config",
    type=Path,
    default=None,
    help="The config file path. (default: ./manager.conf and /etc/backend.ai/manager.conf)",
)
@click.option(
    "--debug",
    is_flag=True,
    help="Enable the debug mode and override the global log level to DEBUG.",
)
@click.pass_context
def main(ctx, config_path, debug):
    """
    Manager Administration CLI
    """
    local_config = load_config(config_path)
    setproctitle(f"backend.ai: manager.cli {local_config['etcd']['namespace']}")
    ctx.obj = CLIContext(
        logger=init_logger(local_config),
        local_config=local_config,
    )


@main.command(
    context_settings=dict(
        ignore_unknown_options=True,
    )
)
@click.option(
    "--psql-container",
    "container_name",
    type=str,
    default=None,
    metavar="ID_OR_NAME",
    help="Open a postgres client shell using the psql executable "
    "shipped with the given postgres container. "
    'If not set or set as an empty string "", it will auto-detect '
    "the psql container from the halfstack. "
    'If set "-", it will use the host-provided psql executable. '
    "You may append additional arguments passed to the psql cli command. "
    "[default: auto-detect from halfstack]",
)
@click.option(
    "--psql-help",
    is_flag=True,
    help="Show the help text of the psql command instead of " "this dbshell command.",
)
@click.argument("psql_args", nargs=-1, type=click.UNPROCESSED)
@click.pass_obj
def dbshell(cli_ctx: CLIContext, container_name, psql_help, psql_args):
    """
    Run the database shell.

    All arguments except `--psql-container` and `--psql-help` are transparently
    forwarded to the psql command.  For instance, you can use `-c` to execute a
    psql/SQL statement on the command line.  Note that you do not have to specify
    connection-related options because the dbshell command fills out them from the
    manager configuration.
    """
    local_config = cli_ctx.local_config
    if psql_help:
        psql_args = ["--help"]
    if not container_name:
        # Try to get the database container name of the halfstack
        candidate_container_names = subprocess.check_output(
            ["docker", "ps", "--format", "{{.Names}}", "--filter", "name=half-db"],
        )
        if not candidate_container_names:
            click.echo(
                "Could not find the halfstack postgres container. "
                "Please set the container name explicitly.",
                err=True,
            )
            sys.exit(ExitCode.FAILURE)
        container_name = candidate_container_names.decode().splitlines()[0].strip()
    elif container_name == "-":
        # Use the host-provided psql command
        cmd = [
            "psql",
            (
                f"postgres://{local_config['db']['user']}:{local_config['db']['password']}"
                f"@{local_config['db']['addr']}/{local_config['db']['name']}"
            ),
            *psql_args,
        ]
        subprocess.call(cmd)
        return
    # Use the container to start the psql client command
    print(f"using the db container {container_name} ...")
    cmd = [
        "docker",
        "exec",
        "-i",
        "-t",
        container_name,
        "psql",
        "-U",
        local_config["db"]["user"],
        "-d",
        local_config["db"]["name"],
        *psql_args,
    ]
    subprocess.call(cmd)


@main.command()
@click.pass_obj
def generate_keypair(cli_ctx: CLIContext):
    """
    Generate a random keypair and print it out to stdout.
    """
    from ..models.keypair import generate_keypair as _gen_keypair

    log.info("generating keypair...")
    ak, sk = _gen_keypair()
    print(f"Access Key: {ak} ({len(ak)} bytes)")
    print(f"Secret Key: {sk} ({len(sk)} bytes)")


@main.command()
@click.option(
    "-r",
    "--retention",
    type=str,
    default="1yr",
    help="The retention limit. e.g., 20d, 1mo, 6mo, 1yr",
)
@click.option(
    "-v",
    "--vacuum-full",
    type=bool,
    default=False,
    help="Reclaim storage occupied by dead tuples."
    "If not set or set False, it will run VACUUM without FULL."
    "If set True, it will run VACUUM FULL."
    "When VACUUM FULL is being processed, the database is locked."
    "[default: False]",
)
@click.pass_obj
def clear_history(cli_ctx: CLIContext, retention, vacuum_full) -> None:
    """
    Delete old records from the kernels table and
    invoke the PostgreSQL's vaccuum operation to clear up the actual disk space.
    """
    import sqlalchemy as sa
    from redis.asyncio import Redis
    from redis.asyncio.client import Pipeline

    from ai.backend.manager.models import kernels
    from ai.backend.manager.models.utils import connect_database

    with cli_ctx.logger:
        today = datetime.now()
        duration = TimeDuration()
        expiration_date = today - duration.check_and_return(retention)

        async def _clear_redis_history():
            try:
                async with connect_database(cli_ctx.local_config) as db:
                    async with db.begin_readonly() as conn:
                        query = (
                            sa.select([kernels.c.id])
                            .select_from(kernels)
                            .where(
                                (kernels.c.terminated_at < expiration_date),
                            )
                        )
                        result = await conn.execute(query)
                        target_kernels = [str(x["id"]) for x in result.all()]

                delete_count = 0
                async with redis_ctx(cli_ctx) as redis_conn_set:

                    async def _build_pipe(
                        r: Redis,
                        kernel_ids: list[str],
                    ) -> Pipeline:
                        pipe = r.pipeline(transaction=False)
                        await pipe.delete(*kernel_ids)
                        return pipe

                    if len(target_kernels) > 0:
                        # Apply chunking to avoid excessive length of command params
                        # and indefinite blocking of the Redis server.
                        for kernel_ids in chunked(target_kernels, 32):
                            results = await redis_helper.execute(
                                redis_conn_set.stat,
                                partial(_build_pipe, kernel_ids=kernel_ids),
                            )
                        # Each DEL command returns the number of keys deleted.
                        delete_count += sum(results)
                        log.info(
                            "Cleaned up {:,} redis statistics records older than {:}.",
                            delete_count,
                            expiration_date,
                        )

                    # Sync and compact the persistent database of Redis
                    redis_config = await redis_helper.execute(
                        redis_conn_set.stat,
                        lambda r: r.config_get("appendonly"),
                    )
                    if redis_config["appendonly"] == "yes":
                        await redis_helper.execute(
                            redis_conn_set.stat,
                            lambda r: r.bgrewriteaof(),
                        )
                        log.info("Issued BGREWRITEAOF to the Redis database.")
                    else:
                        await redis_helper.execute(
                            redis_conn_set.stat,
                            lambda r: r.execute_command("BGSAVE SCHEDULE"),
                        )
                        log.info("Issued BGSAVE to the Redis database.")
            except Exception:
                log.exception("Unexpected error while cleaning up redis history")

        async def _clear_terminated_sessions():
            async with connect_database(cli_ctx.local_config, isolation_level="AUTOCOMMIT") as db:
                async with db.begin() as conn:
                    log.info("Deleting old records...")
                    result = await conn.execute(
                        sa.delete(kernels).where(kernels.c.terminated_at < expiration_date),
                    )
                    deleted_count = result.rowcount

                    vacuum_sql = "VACUUM FULL" if vacuum_full else "VACUUM"
                    log.info(f"Perfoming {vacuum_sql} operation...")
                    await conn.exec_driver_sql(vacuum_sql)

                    curs = await conn.execute(sa.select([sa.func.count()]).select_from(kernels))
                    if ret := curs.fetchone():
                        table_size = ret[0]
                        log.info(
                            "The number of rows of the `kernels` tables after cleanup: {}",
                            table_size,
                        )
            log.info(
                "Cleaned up {:,} database records older than {}.",
                deleted_count,
                expiration_date,
            )

        asyncio.run(_clear_redis_history())
        asyncio.run(_clear_terminated_sessions())


@main.group(cls=LazyGroup, import_name="ai.backend.manager.cli.dbschema:cli")
def schema():
    """Command set for managing the database schema."""


@main.group(cls=LazyGroup, import_name="ai.backend.manager.cli.etcd:cli")
def etcd():
    """Command set for putting/getting data to/from etcd."""


@main.group(cls=LazyGroup, import_name="ai.backend.manager.cli.fixture:cli")
def fixture():
    """Command set for managing fixtures."""


@main.group(cls=LazyGroup, import_name="ai.backend.manager.cli.gql:cli")
def gql():
    """Command set for GraphQL schema."""


@main.group(cls=LazyGroup, import_name="ai.backend.manager.cli.image:cli")
def image():
    """Command set for managing images."""


if __name__ == "__main__":
    main()



import json
import logging
from typing import AsyncIterator, Optional, cast

import aiohttp
import yarl

from ai.backend.common.docker import login as registry_login
from ai.backend.common.logging import BraceStyleAdapter

from .base import BaseContainerRegistry

log = BraceStyleAdapter(logging.getLogger(__name__))


class DockerHubRegistry(BaseContainerRegistry):
    async def fetch_repositories(
        self,
        sess: aiohttp.ClientSession,
    ) -> AsyncIterator[str]:
        # We need some special treatment for the Docker Hub.
        params = {"page_size": "30"}
        username = self.registry_info["username"]
        hub_url = yarl.URL("https://hub.docker.com")
        repo_list_url: Optional[yarl.URL]
        repo_list_url = hub_url / f"v2/repositories/{username}/"
        while repo_list_url is not None:
            async with sess.get(repo_list_url, params=params) as resp:
                if resp.status == 200:
                    data = await resp.json()
                    for item in data["results"]:
                        # skip legacy images
                        if item["name"].startswith("kernel-"):
                            continue
                        yield f"{username}/{item['name']}"
                else:
                    log.error(
                        "Failed to fetch repository list from {0} " "(status={1})",
                        repo_list_url,
                        resp.status,
                    )
                    break
            repo_list_url = None
            next_page_link = data.get("next", None)
            if next_page_link:
                next_page_url = yarl.URL(next_page_link)
                repo_list_url = hub_url.with_path(next_page_url.path).with_query(
                    next_page_url.query
                )


class DockerRegistry_v2(BaseContainerRegistry):
    async def fetch_repositories(
        self,
        sess: aiohttp.ClientSession,
    ) -> AsyncIterator[str]:
        # The credential should have the catalog search privilege.
        rqst_args = await registry_login(
            sess,
            self.registry_url,
            self.credentials,
            "registry:catalog:*",
        )
        catalog_url: Optional[yarl.URL]
        catalog_url = (self.registry_url / "v2/_catalog").with_query(
            {"n": "30"},
        )
        while catalog_url is not None:
            async with sess.get(catalog_url, **rqst_args) as resp:
                if resp.status == 200:
                    data = json.loads(await resp.read())
                    for item in data["repositories"]:
                        yield item
                    log.debug("found {} repositories", len(data["repositories"]))
                else:
                    log.warning(
                        "Docker registry {0} does not allow/support "
                        "catalog search. (status={1})",
                        self.registry_url,
                        resp.status,
                    )
                    break
                catalog_url = None
                next_page_link = resp.links.get("next")
                if next_page_link:
                    next_page_url = cast(yarl.URL, next_page_link["url"])
                    catalog_url = self.registry_url.with_path(next_page_url.path).with_query(
                        next_page_url.query
                    )



from __future__ import annotations

import asyncio
import functools
import grp
import importlib
import logging
import os
import pwd
import ssl
import sys
import traceback
from contextlib import asynccontextmanager as actxmgr
from datetime import datetime
from pathlib import Path
from typing import (
    Any,
    AsyncIterator,
    Final,
    Iterable,
    List,
    Mapping,
    MutableMapping,
    Sequence,
    cast,
)

import aiohttp_cors
import aiomonitor
import aiotools
import click
from aiohttp import web
from redis.asyncio import Redis
from setproctitle import setproctitle

from ai.backend.common import redis_helper
from ai.backend.common.bgtask import BackgroundTaskManager
from ai.backend.common.cli import LazyGroup
from ai.backend.common.events import EventDispatcher, EventProducer
from ai.backend.common.logging import BraceStyleAdapter, Logger
from ai.backend.common.plugin.hook import ALL_COMPLETED, PASSED, HookPluginContext
from ai.backend.common.plugin.monitor import INCREMENT
from ai.backend.common.utils import env_info

from . import __version__
from .api import ManagerStatus
from .api.context import RootContext
from .api.exceptions import (
    BackendError,
    GenericBadRequest,
    InternalServerError,
    InvalidAPIParameters,
    MethodNotAllowed,
    URLNotFound,
)
from .api.types import AppCreator, CleanupContext, WebMiddleware, WebRequestHandler
from .config import LocalConfig, SharedConfig
from .config import load as load_config
from .config import volume_config_iv
from .defs import REDIS_IMAGE_DB, REDIS_LIVE_DB, REDIS_STAT_DB, REDIS_STREAM_DB, REDIS_STREAM_LOCK
from .exceptions import InvalidArgument
from .types import DistributedLockFactory

VALID_VERSIONS: Final = frozenset(
    [
        # 'v1.20160915',  # deprecated
        # 'v2.20170315',  # deprecated
        # 'v3.20170615',  # deprecated
        # authentication changed not to use request bodies
        "v4.20181215",
        # added & enabled streaming-execute API
        "v4.20190115",
        # changed resource/image formats
        "v4.20190315",
        # added user mgmt and ID/password authentication
        # added domain/group/scaling-group
        # added domain/group/scaling-group ref. fields to user/keypair/vfolder objects
        "v4.20190615",
        # added mount_map parameter when creating kernel
        # changed GraphQL query structures for multi-container bundled sessions
        "v5.20191215",
        # rewrote vfolder upload/download APIs to migrate to external storage proxies
        "v6.20200815",
        # added standard-compliant /admin/gql endpoint
        # deprecated /admin/graphql endpoint (still present for backward compatibility)
        # added "groups_by_name" GQL query
        # added "filter" and "order" arg to all paginated GQL queries with their own expression mini-langs
        # removed "order_key" and "order_asc" arguments from all paginated GQL queries (never used!)
        "v6.20210815",
        # added session dependencies and state callback URLs configs when creating sessions
        # added session event webhook option to session creation API
        # added architecture option when making image aliases
        "v6.20220315",
        # added payload encryption / decryption on selected transfer
        "v6.20220615",
    ]
)
LATEST_REV_DATES: Final = {
    1: "20160915",
    2: "20170915",
    3: "20181215",
    4: "20190615",
    5: "20191215",
    6: "20220615",
}
LATEST_API_VERSION: Final = "v6.20220615"

log = BraceStyleAdapter(logging.getLogger(__name__))

PUBLIC_INTERFACES: Final = [
    "pidx",
    "background_task_manager",
    "local_config",
    "shared_config",
    "db",
    "registry",
    "redis_live",
    "redis_stat",
    "redis_image",
    "redis_stream",
    "event_dispatcher",
    "event_producer",
    "idle_checkers",
    "storage_manager",
    "stats_monitor",
    "error_monitor",
    "hook_plugin_ctx",
]

public_interface_objs: MutableMapping[str, Any] = {}


async def hello(request: web.Request) -> web.Response:
    """
    Returns the API version number.
    """
    return web.json_response(
        {
            "version": LATEST_API_VERSION,
            "manager": __version__,
        }
    )


async def on_prepare(request: web.Request, response: web.StreamResponse) -> None:
    response.headers["Server"] = "BackendAI"


@web.middleware
async def api_middleware(request: web.Request, handler: WebRequestHandler) -> web.StreamResponse:
    _handler = handler
    method_override = request.headers.get("X-Method-Override", None)
    if method_override:
        request = request.clone(method=method_override)
        new_match_info = await request.app.router.resolve(request)
        if new_match_info is None:
            raise InternalServerError("No matching method handler found")
        _handler = new_match_info.handler
        request._match_info = new_match_info  # type: ignore  # this is a hack
    ex = request.match_info.http_exception
    if ex is not None:
        # handled by exception_middleware
        raise ex
    new_api_version = request.headers.get("X-BackendAI-Version")
    legacy_api_version = request.headers.get("X-Sorna-Version")
    api_version = new_api_version or legacy_api_version
    try:
        if api_version is None:
            path_major_version = int(request.match_info.get("version", 5))
            revision_date = LATEST_REV_DATES[path_major_version]
            request["api_version"] = (path_major_version, revision_date)
        elif api_version in VALID_VERSIONS:
            hdr_major_version, revision_date = api_version.split(".", maxsplit=1)
            request["api_version"] = (int(hdr_major_version[1:]), revision_date)
        else:
            return GenericBadRequest("Unsupported API version.")
    except (ValueError, KeyError):
        return GenericBadRequest("Unsupported API version.")
    resp = await _handler(request)
    return resp


@web.middleware
async def exception_middleware(
    request: web.Request, handler: WebRequestHandler
) -> web.StreamResponse:
    root_ctx: RootContext = request.app["_root.context"]
    error_monitor = root_ctx.error_monitor
    stats_monitor = root_ctx.stats_monitor
    try:
        await stats_monitor.report_metric(INCREMENT, "ai.backend.manager.api.requests")
        resp = await handler(request)
    except InvalidArgument as ex:
        if len(ex.args) > 1:
            raise InvalidAPIParameters(f"{ex.args[0]}: {', '.join(map(str, ex.args[1:]))}")
        elif len(ex.args) == 1:
            raise InvalidAPIParameters(ex.args[0])
        else:
            raise InvalidAPIParameters()
    except BackendError as ex:
        if ex.status_code == 500:
            log.warning("Internal server error raised inside handlers")
        await error_monitor.capture_exception()
        await stats_monitor.report_metric(INCREMENT, "ai.backend.manager.api.failures")
        await stats_monitor.report_metric(
            INCREMENT, f"ai.backend.manager.api.status.{ex.status_code}"
        )
        raise
    except web.HTTPException as ex:
        await stats_monitor.report_metric(INCREMENT, "ai.backend.manager.api.failures")
        await stats_monitor.report_metric(
            INCREMENT, f"ai.backend.manager.api.status.{ex.status_code}"
        )
        if ex.status_code == 404:
            raise URLNotFound(extra_data=request.path)
        if ex.status_code == 405:
            concrete_ex = cast(web.HTTPMethodNotAllowed, ex)
            raise MethodNotAllowed(concrete_ex.method, concrete_ex.allowed_methods)
        log.warning("Bad request: {0!r}", ex)
        raise GenericBadRequest
    except asyncio.CancelledError as e:
        # The server is closing or the client has disconnected in the middle of
        # request.  Atomic requests are still executed to their ends.
        log.debug("Request cancelled ({0} {1})", request.method, request.rel_url)
        raise e
    except Exception as e:
        await error_monitor.capture_exception()
        log.exception("Uncaught exception in HTTP request handlers {0!r}", e)
        if root_ctx.local_config["debug"]["enabled"]:
            raise InternalServerError(traceback.format_exc())
        else:
            raise InternalServerError()
    else:
        await stats_monitor.report_metric(INCREMENT, f"ai.backend.manager.api.status.{resp.status}")
        return resp


@actxmgr
async def shared_config_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    # populate public interfaces
    root_ctx.shared_config = SharedConfig(
        root_ctx.local_config["etcd"]["addr"],
        root_ctx.local_config["etcd"]["user"],
        root_ctx.local_config["etcd"]["password"],
        root_ctx.local_config["etcd"]["namespace"],
    )
    await root_ctx.shared_config.reload()
    yield
    await root_ctx.shared_config.close()


@actxmgr
async def webapp_plugin_ctx(root_app: web.Application) -> AsyncIterator[None]:
    from .plugin.webapp import WebappPluginContext

    root_ctx: RootContext = root_app["_root.context"]
    plugin_ctx = WebappPluginContext(root_ctx.shared_config.etcd, root_ctx.local_config)
    await plugin_ctx.init(
        allowlist=root_ctx.local_config["manager"]["allowed-plugins"],
        blocklist=root_ctx.local_config["manager"]["disabled-plugins"],
    )
    root_ctx.webapp_plugin_ctx = plugin_ctx
    for plugin_name, plugin_instance in plugin_ctx.plugins.items():
        if root_ctx.pidx == 0:
            log.info("Loading webapp plugin: {0}", plugin_name)
        subapp, global_middlewares = await plugin_instance.create_app(root_ctx.cors_options)
        _init_subapp(plugin_name, root_app, subapp, global_middlewares)
    yield
    await plugin_ctx.cleanup()


@actxmgr
async def manager_status_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    if root_ctx.pidx == 0:
        mgr_status = await root_ctx.shared_config.get_manager_status()
        if mgr_status is None or mgr_status not in (ManagerStatus.RUNNING, ManagerStatus.FROZEN):
            # legacy transition: we now have only RUNNING or FROZEN for HA setup.
            await root_ctx.shared_config.update_manager_status(ManagerStatus.RUNNING)
            mgr_status = ManagerStatus.RUNNING
        log.info("Manager status: {}", mgr_status)
        tz = root_ctx.shared_config["system"]["timezone"]
        log.info("Configured timezone: {}", tz.tzname(datetime.now()))
    yield


@actxmgr
async def redis_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    root_ctx.redis_live = redis_helper.get_redis_object(
        root_ctx.shared_config.data["redis"],
        db=REDIS_LIVE_DB,
    )
    root_ctx.redis_stat = redis_helper.get_redis_object(
        root_ctx.shared_config.data["redis"],
        db=REDIS_STAT_DB,
    )
    root_ctx.redis_image = redis_helper.get_redis_object(
        root_ctx.shared_config.data["redis"],
        db=REDIS_IMAGE_DB,
    )
    root_ctx.redis_stream = redis_helper.get_redis_object(
        root_ctx.shared_config.data["redis"],
        db=REDIS_STREAM_DB,
    )
    root_ctx.redis_lock = redis_helper.get_redis_object(
        root_ctx.shared_config.data["redis"],
        db=REDIS_STREAM_LOCK,
    )
    for redis_info in (
        root_ctx.redis_live,
        root_ctx.redis_stat,
        root_ctx.redis_image,
        root_ctx.redis_stream,
        root_ctx.redis_lock,
    ):
        assert isinstance(redis_info.client, Redis)
        await redis_helper.ping_redis_connection(redis_info.client)
    yield
    await root_ctx.redis_stream.close()
    await root_ctx.redis_image.close()
    await root_ctx.redis_stat.close()
    await root_ctx.redis_live.close()
    await root_ctx.redis_lock.close()


@actxmgr
async def database_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    from .models.utils import connect_database

    async with connect_database(root_ctx.local_config) as db:
        root_ctx.db = db
        yield


@actxmgr
async def distributed_lock_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    root_ctx.distributed_lock_factory = init_lock_factory(root_ctx)
    yield


@actxmgr
async def event_dispatcher_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    root_ctx.event_producer = await EventProducer.new(
        root_ctx.shared_config.data["redis"],
        db=REDIS_STREAM_DB,
    )
    root_ctx.event_dispatcher = await EventDispatcher.new(
        root_ctx.shared_config.data["redis"],
        db=REDIS_STREAM_DB,
        log_events=root_ctx.local_config["debug"]["log-events"],
        node_id=root_ctx.local_config["manager"]["id"],
    )
    yield
    await root_ctx.event_producer.close()
    await asyncio.sleep(0.2)
    await root_ctx.event_dispatcher.close()


@actxmgr
async def idle_checker_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    from .idle import init_idle_checkers

    root_ctx.idle_checker_host = await init_idle_checkers(
        root_ctx.db,
        root_ctx.shared_config,
        root_ctx.event_dispatcher,
        root_ctx.event_producer,
        root_ctx.distributed_lock_factory,
    )
    await root_ctx.idle_checker_host.start()
    yield
    await root_ctx.idle_checker_host.shutdown()


@actxmgr
async def storage_manager_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    from .models.storage import StorageSessionManager

    raw_vol_config = await root_ctx.shared_config.etcd.get_prefix("volumes")
    config = volume_config_iv.check(raw_vol_config)
    root_ctx.storage_manager = StorageSessionManager(config)
    yield
    await root_ctx.storage_manager.aclose()


@actxmgr
async def hook_plugin_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    ctx = HookPluginContext(root_ctx.shared_config.etcd, root_ctx.local_config)
    root_ctx.hook_plugin_ctx = ctx
    await ctx.init(
        allowlist=root_ctx.local_config["manager"]["allowed-plugins"],
        blocklist=root_ctx.local_config["manager"]["disabled-plugins"],
    )
    hook_result = await ctx.dispatch(
        "ACTIVATE_MANAGER",
        (),
        return_when=ALL_COMPLETED,
    )
    if hook_result.status != PASSED:
        raise RuntimeError("Could not activate the manager instance.")
    yield
    await ctx.cleanup()


@actxmgr
async def agent_registry_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    from .registry import AgentRegistry

    root_ctx.registry = AgentRegistry(
        root_ctx.shared_config,
        root_ctx.db,
        root_ctx.redis_stat,
        root_ctx.redis_live,
        root_ctx.redis_image,
        root_ctx.event_dispatcher,
        root_ctx.event_producer,
        root_ctx.storage_manager,
        root_ctx.hook_plugin_ctx,
    )
    await root_ctx.registry.init()
    yield
    await root_ctx.registry.shutdown()


@actxmgr
async def sched_dispatcher_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    from .scheduler.dispatcher import SchedulerDispatcher

    sched_dispatcher = await SchedulerDispatcher.new(
        root_ctx.local_config,
        root_ctx.shared_config,
        root_ctx.event_dispatcher,
        root_ctx.event_producer,
        root_ctx.distributed_lock_factory,
        root_ctx.registry,
    )
    yield
    await sched_dispatcher.close()


@actxmgr
async def monitoring_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    from .plugin.monitor import ManagerErrorPluginContext, ManagerStatsPluginContext

    ectx = ManagerErrorPluginContext(root_ctx.shared_config.etcd, root_ctx.local_config)
    sctx = ManagerStatsPluginContext(root_ctx.shared_config.etcd, root_ctx.local_config)
    init_success = False
    try:
        await ectx.init(
            context={"_root.context": root_ctx},
            allowlist=root_ctx.local_config["manager"]["allowed-plugins"],
        )
        await sctx.init(allowlist=root_ctx.local_config["manager"]["allowed-plugins"])
    except Exception:
        log.error("Failed to initialize monitoring plugins")
    else:
        init_success = True
        root_ctx.error_monitor = ectx
        root_ctx.stats_monitor = sctx
    yield
    if init_success:
        await sctx.cleanup()
        await ectx.cleanup()


class background_task_ctx:
    def __init__(self, root_ctx: RootContext) -> None:
        self.root_ctx = root_ctx

    async def __aenter__(self) -> None:
        self.root_ctx.background_task_manager = BackgroundTaskManager(self.root_ctx.event_producer)

    async def __aexit__(self, *exc_info) -> None:
        pass

    async def shutdown(self) -> None:
        if hasattr(self.root_ctx, "background_task_manager"):
            await self.root_ctx.background_task_manager.shutdown()


def handle_loop_error(
    root_ctx: RootContext,
    loop: asyncio.AbstractEventLoop,
    context: Mapping[str, Any],
) -> None:
    exception = context.get("exception")
    msg = context.get("message", "(empty message)")
    if exception is not None:
        if sys.exc_info()[0] is not None:
            log.exception("Error inside event loop: {0}", msg)
            if (error_monitor := getattr(root_ctx, "error_monitor", None)) is not None:
                loop.create_task(error_monitor.capture_exception())
        else:
            exc_info = (type(exception), exception, exception.__traceback__)
            log.error("Error inside event loop: {0}", msg, exc_info=exc_info)
            if (error_monitor := getattr(root_ctx, "error_monitor", None)) is not None:
                loop.create_task(error_monitor.capture_exception(exc_instance=exception))


def _init_subapp(
    pkg_name: str,
    root_app: web.Application,
    subapp: web.Application,
    global_middlewares: Iterable[WebMiddleware],
) -> None:
    subapp.on_response_prepare.append(on_prepare)

    async def _set_root_ctx(subapp: web.Application):
        # Allow subapp's access to the root app properties.
        # These are the public APIs exposed to plugins as well.
        subapp["_root.context"] = root_app["_root.context"]

    # We must copy the public interface prior to all user-defined startup signal handlers.
    subapp.on_startup.insert(0, _set_root_ctx)
    prefix = subapp.get("prefix", pkg_name.split(".")[-1].replace("_", "-"))
    root_app.add_subapp("/" + prefix, subapp)
    root_app.middlewares.extend(global_middlewares)


def init_subapp(pkg_name: str, root_app: web.Application, create_subapp: AppCreator) -> None:
    root_ctx: RootContext = root_app["_root.context"]
    subapp, global_middlewares = create_subapp(root_ctx.cors_options)
    _init_subapp(pkg_name, root_app, subapp, global_middlewares)


def init_lock_factory(root_ctx: RootContext) -> DistributedLockFactory:
    ipc_base_path = root_ctx.local_config["manager"]["ipc-base-path"]
    manager_id = root_ctx.local_config["manager"]["id"]
    lock_backend = root_ctx.local_config["manager"]["distributed-lock"]
    log.debug("using {} as the distributed lock backend", lock_backend)
    match lock_backend:
        case "filelock":
            from ai.backend.common.lock import FileLock

            return lambda lock_id, lifetime_hint: FileLock(
                ipc_base_path / f"{manager_id}.{lock_id}.lock",
                timeout=0,
            )
        case "pg_advisory":
            from .pglock import PgAdvisoryLock

            return lambda lock_id, lifetime_hint: PgAdvisoryLock(root_ctx.db, lock_id)
        case "redlock":
            from ai.backend.common.lock import RedisLock

            return lambda lock_id, lifetime_hint: RedisLock(
                str(lock_id),
                root_ctx.redis_lock,
                lifetime=min(lifetime_hint * 2, lifetime_hint + 30),
            )
        case "etcd":
            from ai.backend.common.lock import EtcdLock

            return lambda lock_id, lifetime_hint: EtcdLock(
                str(lock_id),
                root_ctx.shared_config.etcd,
                lifetime=min(lifetime_hint * 2, lifetime_hint + 30),
            )
        case other:
            raise ValueError(f"Invalid lock backend: {other}")


def build_root_app(
    pidx: int,
    local_config: LocalConfig,
    *,
    cleanup_contexts: Sequence[CleanupContext] = None,
    subapp_pkgs: Sequence[str] = None,
    scheduler_opts: Mapping[str, Any] = None,
) -> web.Application:
    public_interface_objs.clear()
    app = web.Application(
        middlewares=[
            exception_middleware,
            api_middleware,
        ]
    )
    root_ctx = RootContext()
    global_exception_handler = functools.partial(handle_loop_error, root_ctx)
    loop = asyncio.get_running_loop()
    loop.set_exception_handler(global_exception_handler)
    app["_root.context"] = root_ctx
    root_ctx.local_config = local_config
    root_ctx.pidx = pidx
    root_ctx.cors_options = {
        "*": aiohttp_cors.ResourceOptions(
            allow_credentials=False, expose_headers="*", allow_headers="*"
        ),
    }
    default_scheduler_opts = {
        "limit": 2048,
        "close_timeout": 30,
        "exception_handler": global_exception_handler,
    }
    app["scheduler_opts"] = {
        **default_scheduler_opts,
        **(scheduler_opts if scheduler_opts is not None else {}),
    }
    app.on_response_prepare.append(on_prepare)

    if cleanup_contexts is None:
        cleanup_contexts = [
            manager_status_ctx,
            redis_ctx,
            database_ctx,
            distributed_lock_ctx,
            event_dispatcher_ctx,
            idle_checker_ctx,
            storage_manager_ctx,
            hook_plugin_ctx,
            monitoring_ctx,
            agent_registry_ctx,
            sched_dispatcher_ctx,
            background_task_ctx,
        ]

    async def _cleanup_context_wrapper(cctx, app: web.Application) -> AsyncIterator[None]:
        # aiohttp's cleanup contexts are just async generators, not async context managers.
        cctx_instance = cctx(app["_root.context"])
        app["_cctx_instances"].append(cctx_instance)
        try:
            async with cctx_instance:
                yield
        except Exception as e:
            exc_info = (type(e), e, e.__traceback__)
            log.error("Error initializing cleanup_contexts: {0}", cctx.__name__, exc_info=exc_info)

    async def _call_cleanup_context_shutdown_handlers(app: web.Application) -> None:
        for cctx in app["_cctx_instances"]:
            if hasattr(cctx, "shutdown"):
                try:
                    await cctx.shutdown()
                except Exception:
                    log.exception("error while shutting down a cleanup context")

    app["_cctx_instances"] = []
    app.on_shutdown.append(_call_cleanup_context_shutdown_handlers)
    for cleanup_ctx in cleanup_contexts:
        app.cleanup_ctx.append(
            functools.partial(_cleanup_context_wrapper, cleanup_ctx),
        )
    cors = aiohttp_cors.setup(app, defaults=root_ctx.cors_options)
    # should be done in create_app() in other modules.
    cors.add(app.router.add_route("GET", r"", hello))
    cors.add(app.router.add_route("GET", r"/", hello))
    if subapp_pkgs is None:
        subapp_pkgs = []
    for pkg_name in subapp_pkgs:
        if pidx == 0:
            log.info("Loading module: {0}", pkg_name[1:])
        subapp_mod = importlib.import_module(pkg_name, "ai.backend.manager.api")
        init_subapp(pkg_name, app, getattr(subapp_mod, "create_app"))
    return app


@actxmgr
async def server_main(
    loop: asyncio.AbstractEventLoop,
    pidx: int,
    _args: List[Any],
) -> AsyncIterator[None]:
    subapp_pkgs = [
        ".acl",
        ".etcd",
        ".events",
        ".auth",
        ".ratelimit",
        ".vfolder",
        ".admin",
        ".session",
        ".stream",
        ".manager",
        ".resource",
        ".scaling_group",
        ".cluster_template",
        ".session_template",
        ".image",
        ".userconfig",
        ".domainconfig",
        ".groupconfig",
        ".logs",
    ]
    root_app = build_root_app(pidx, _args[0], subapp_pkgs=subapp_pkgs)
    root_ctx: RootContext = root_app["_root.context"]

    # Start aiomonitor.
    # Port is set by config (default=50100 + pidx).
    loop.set_debug(root_ctx.local_config["debug"]["asyncio"])
    m = aiomonitor.Monitor(
        loop,
        port=root_ctx.local_config["manager"]["aiomonitor-port"] + pidx,
        console_enabled=False,
        hook_task_factory=root_ctx.local_config["debug"]["enhanced-aiomonitor-task-info"],
    )
    m.prompt = f"monitor (manager[{pidx}@{os.getpid()}]) >>> "
    # Add some useful console_locals for ease of debugging
    m.console_locals["root_app"] = root_app
    m.console_locals["root_ctx"] = root_ctx
    aiomon_started = False
    try:
        m.start()
        aiomon_started = True
    except Exception as e:
        log.warning("aiomonitor could not start but skipping this error to continue", exc_info=e)

    # Plugin webapps should be loaded before runner.setup(),
    # which freezes on_startup event.
    try:
        async with (
            shared_config_ctx(root_ctx),
            webapp_plugin_ctx(root_app),
        ):
            ssl_ctx = None
            if root_ctx.local_config["manager"]["ssl-enabled"]:
                ssl_ctx = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
                ssl_ctx.load_cert_chain(
                    str(root_ctx.local_config["manager"]["ssl-cert"]),
                    str(root_ctx.local_config["manager"]["ssl-privkey"]),
                )

            runner = web.AppRunner(root_app, keepalive_timeout=30.0)
            await runner.setup()
            service_addr = root_ctx.local_config["manager"]["service-addr"]
            site = web.TCPSite(
                runner,
                str(service_addr.host),
                service_addr.port,
                backlog=1024,
                reuse_port=True,
                ssl_context=ssl_ctx,
            )
            await site.start()

            if os.geteuid() == 0:
                uid = root_ctx.local_config["manager"]["user"]
                gid = root_ctx.local_config["manager"]["group"]
                os.setgroups(
                    [g.gr_gid for g in grp.getgrall() if pwd.getpwuid(uid).pw_name in g.gr_mem]
                )
                os.setgid(gid)
                os.setuid(uid)
                log.info("changed process uid and gid to {}:{}", uid, gid)
            log.info("started handling API requests at {}", service_addr)

            try:
                yield
            finally:
                log.info("shutting down...")
                await runner.cleanup()
    finally:
        if aiomon_started:
            m.close()


@actxmgr
async def server_main_logwrapper(
    loop: asyncio.AbstractEventLoop,
    pidx: int,
    _args: List[Any],
) -> AsyncIterator[None]:
    setproctitle(f"backend.ai: manager worker-{pidx}")
    log_endpoint = _args[1]
    logger = Logger(_args[0]["logging"], is_master=False, log_endpoint=log_endpoint)
    try:
        with logger:
            async with server_main(loop, pidx, _args):
                yield
    except Exception:
        traceback.print_exc()


@click.group(invoke_without_command=True)
@click.option(
    "-f",
    "--config-path",
    "--config",
    type=Path,
    default=None,
    help="The config file path. (default: ./manager.toml and /etc/backend.ai/manager.toml)",
)
@click.option(
    "--debug",
    is_flag=True,
    help="Enable the debug mode and override the global log level to DEBUG.",
)
@click.pass_context
def main(ctx: click.Context, config_path: Path, debug: bool) -> None:
    """
    Start the manager service as a foreground process.
    """

    cfg = load_config(config_path, debug)

    if ctx.invoked_subcommand is None:
        cfg["manager"]["pid-file"].write_text(str(os.getpid()))
        ipc_base_path = cfg["manager"]["ipc-base-path"]
        log_sockpath = ipc_base_path / f"manager-logger-{os.getpid()}.sock"
        log_endpoint = f"ipc://{log_sockpath}"
        try:
            logger = Logger(cfg["logging"], is_master=True, log_endpoint=log_endpoint)
            with logger:
                ns = cfg["etcd"]["namespace"]
                setproctitle(f"backend.ai: manager {ns}")
                log.info("Backend.AI Manager {0}", __version__)
                log.info("runtime: {0}", env_info())
                log_config = logging.getLogger("ai.backend.manager.config")
                log_config.debug("debug mode enabled.")
                if cfg["manager"]["event-loop"] == "uvloop":
                    import uvloop

                    uvloop.install()
                    log.info("Using uvloop as the event loop backend")
                try:
                    aiotools.start_server(
                        server_main_logwrapper,
                        num_workers=cfg["manager"]["num-proc"],
                        args=(cfg, log_endpoint),
                        wait_timeout=5.0,
                    )
                finally:
                    log.info("terminated.")
        finally:
            if cfg["manager"]["pid-file"].is_file():
                # check is_file() to prevent deleting /dev/null!
                cfg["manager"]["pid-file"].unlink()
    else:
        # Click is going to invoke a subcommand.
        pass


@main.group(cls=LazyGroup, import_name="ai.backend.manager.api.auth:cli")
def auth() -> None:
    pass


if __name__ == "__main__":
    sys.exit(main())



"""add state column to vfolder_invitations

Revision ID: f9971fbb34d9
Revises: 185852ff9872
Create Date: 2018-07-12 23:30:14.942845

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "f9971fbb34d9"
down_revision = "185852ff9872"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "vfolder_invitations",
        sa.Column(
            "created_at", sa.DateTime(timezone=True), server_default=sa.text("now()"), nullable=True
        ),
    )
    op.add_column("vfolder_invitations", sa.Column("state", sa.String(length=10), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("vfolder_invitations", "state")
    op.drop_column("vfolder_invitations", "created_at")
    # ### end Alembic commands ###



"""agent_add_region

Revision ID: 93e9d31d40bf
Revises: 80176413d8aa
Create Date: 2017-09-28 15:01:38.944738

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "93e9d31d40bf"
down_revision = "80176413d8aa"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "agents",
        sa.Column(
            "region", sa.String(length=64), nullable=False, server_default="amazon/ap-northeast-2"
        ),
    )
    op.create_index(op.f("ix_agents_region"), "agents", ["region"], unique=False)
    op.alter_column(
        "keypairs",
        "is_admin",
        existing_type=sa.BOOLEAN(),
        nullable=True,
        existing_server_default=sa.text("false"),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column(
        "keypairs",
        "is_admin",
        existing_type=sa.BOOLEAN(),
        nullable=False,
        existing_server_default=sa.text("false"),
    )
    op.drop_index(op.f("ix_agents_region"), table_name="agents")
    op.drop_column("agents", "region")
    # ### end Alembic commands ###



from __future__ import annotations

import enum
import os.path
import uuid
from pathlib import PurePosixPath
from typing import TYPE_CHECKING, Any, List, Mapping, Optional, Sequence

import graphene
import sqlalchemy as sa
import trafaret as t
from dateutil.parser import parse as dtparse
from graphene.types.datetime import DateTime as GQLDateTime
from sqlalchemy.engine.row import Row
from sqlalchemy.ext.asyncio import AsyncConnection as SAConnection

from ai.backend.common.types import VFolderHostPermissionMap, VFolderMount

from ..api.exceptions import InvalidAPIParameters, VFolderNotFound, VFolderOperationFailed
from ..defs import RESERVED_VFOLDER_PATTERNS, RESERVED_VFOLDERS
from ..types import UserScope
from .base import (
    GUID,
    BigInt,
    EnumValueType,
    IDColumn,
    Item,
    PaginatedList,
    batch_multiresult,
    metadata,
)
from .minilang.ordering import QueryOrderParser
from .minilang.queryfilter import QueryFilterParser
from .user import UserRole

if TYPE_CHECKING:
    from .gql import GraphQueryContext
    from .storage import StorageSessionManager

__all__: Sequence[str] = (
    "vfolders",
    "vfolder_invitations",
    "vfolder_permissions",
    "VirtualFolder",
    "VFolderUsageMode",
    "VFolderOwnershipType",
    "VFolderInvitationState",
    "VFolderPermission",
    "VFolderPermissionValidator",
    "VFolderOperationStatus",
    "VFolderAccessStatus",
    "query_accessible_vfolders",
    "get_allowed_vfolder_hosts_by_group",
    "get_allowed_vfolder_hosts_by_user",
    "verify_vfolder_name",
    "prepare_vfolder_mounts",
)


class VFolderUsageMode(str, enum.Enum):
    """
    Usage mode of virtual folder.

    GENERAL: normal virtual folder
    MODEL: virtual folder which provides shared models
    DATA: virtual folder which provides shared data
    """

    GENERAL = "general"
    MODEL = "model"
    DATA = "data"


class VFolderOwnershipType(str, enum.Enum):
    """
    Ownership type of virtual folder.
    """

    USER = "user"
    GROUP = "group"


class VFolderPermission(str, enum.Enum):
    """
    Permissions for a virtual folder given to a specific access key.
    RW_DELETE includes READ_WRITE and READ_WRITE includes READ_ONLY.
    """

    READ_ONLY = "ro"
    READ_WRITE = "rw"
    RW_DELETE = "wd"
    OWNER_PERM = "wd"  # resolved as RW_DELETE


class VFolderPermissionValidator(t.Trafaret):
    def check_and_return(self, value: Any) -> VFolderPermission:
        if value not in ["ro", "rw", "wd"]:
            self._failure('one of "ro", "rw", or "wd" required', value=value)
        return VFolderPermission(value)


class VFolderInvitationState(str, enum.Enum):
    """
    Virtual Folder invitation state.
    """

    PENDING = "pending"
    CANCELED = "canceled"  # canceled by inviter
    ACCEPTED = "accepted"
    REJECTED = "rejected"  # rejected by invitee


class VFolderOperationStatus(str, enum.Enum):
    """
    Introduce virtual folder current status for storage-proxy operations.
    """

    READY = "ready"
    PERFORMING = "performing"
    CLONING = "cloning"
    DELETING = "deleting"
    MOUNTED = "mounted"


class VFolderAccessStatus(str, enum.Enum):
    """
    Introduce virtual folder desired status for storage-proxy operations.
    Not added to db scheme  and determined only by current vfolder status.
    """

    READABLE = "readable"
    UPDATABLE = "updatable"


vfolders = sa.Table(
    "vfolders",
    metadata,
    IDColumn("id"),
    # host will be '' if vFolder is unmanaged
    sa.Column("host", sa.String(length=128), nullable=False),
    sa.Column("name", sa.String(length=64), nullable=False, index=True),
    sa.Column(
        "usage_mode",
        EnumValueType(VFolderUsageMode),
        default=VFolderUsageMode.GENERAL,
        nullable=False,
    ),
    sa.Column("permission", EnumValueType(VFolderPermission), default=VFolderPermission.READ_WRITE),
    sa.Column("max_files", sa.Integer(), default=1000),
    sa.Column("max_size", sa.Integer(), default=None),  # in MBytes
    sa.Column("num_files", sa.Integer(), default=0),
    sa.Column("cur_size", sa.Integer(), default=0),  # in KBytes
    sa.Column("created_at", sa.DateTime(timezone=True), server_default=sa.func.now()),
    sa.Column("last_used", sa.DateTime(timezone=True), nullable=True),
    # creator is always set to the user who created vfolder (regardless user/project types)
    sa.Column("creator", sa.String(length=128), nullable=True),
    # unmanaged vfolder represents the host-side absolute path instead of storage-based path.
    sa.Column("unmanaged_path", sa.String(length=512), nullable=True),
    sa.Column(
        "ownership_type",
        EnumValueType(VFolderOwnershipType),
        default=VFolderOwnershipType.USER,
        nullable=False,
    ),
    sa.Column("user", GUID, sa.ForeignKey("users.uuid"), nullable=True),  # owner if user vfolder
    sa.Column("group", GUID, sa.ForeignKey("groups.id"), nullable=True),  # owner if project vfolder
    sa.Column("cloneable", sa.Boolean, default=False, nullable=False),
    sa.Column(
        "status",
        EnumValueType(VFolderOperationStatus),
        default=VFolderOperationStatus.READY,
        server_default=VFolderOperationStatus.READY.value,
        nullable=False,
    ),
    sa.CheckConstraint(
        "(ownership_type = 'user' AND \"user\" IS NOT NULL) OR "
        "(ownership_type = 'group' AND \"group\" IS NOT NULL)",
        name="ownership_type_match_with_user_or_group",
    ),
    sa.CheckConstraint(
        '("user" IS NULL AND "group" IS NOT NULL) OR ("user" IS NOT NULL AND "group" IS NULL)',
        name="either_one_of_user_or_group",
    ),
)


vfolder_attachment = sa.Table(
    "vfolder_attachment",
    metadata,
    sa.Column(
        "vfolder",
        GUID,
        sa.ForeignKey("vfolders.id", onupdate="CASCADE", ondelete="CASCADE"),
        nullable=False,
    ),
    sa.Column(
        "kernel",
        GUID,
        sa.ForeignKey("kernels.id", onupdate="CASCADE", ondelete="CASCADE"),
        nullable=False,
    ),
    sa.PrimaryKeyConstraint("vfolder", "kernel"),
)


vfolder_invitations = sa.Table(
    "vfolder_invitations",
    metadata,
    IDColumn("id"),
    sa.Column("permission", EnumValueType(VFolderPermission), default=VFolderPermission.READ_WRITE),
    sa.Column("inviter", sa.String(length=256)),  # email
    sa.Column("invitee", sa.String(length=256), nullable=False),  # email
    sa.Column(
        "state", EnumValueType(VFolderInvitationState), default=VFolderInvitationState.PENDING
    ),
    sa.Column("created_at", sa.DateTime(timezone=True), server_default=sa.func.now()),
    sa.Column(
        "modified_at",
        sa.DateTime(timezone=True),
        nullable=True,
        onupdate=sa.func.current_timestamp(),
    ),
    sa.Column(
        "vfolder",
        GUID,
        sa.ForeignKey("vfolders.id", onupdate="CASCADE", ondelete="CASCADE"),
        nullable=False,
    ),
)


vfolder_permissions = sa.Table(
    "vfolder_permissions",
    metadata,
    sa.Column("permission", EnumValueType(VFolderPermission), default=VFolderPermission.READ_WRITE),
    sa.Column(
        "vfolder",
        GUID,
        sa.ForeignKey("vfolders.id", onupdate="CASCADE", ondelete="CASCADE"),
        nullable=False,
    ),
    sa.Column("user", GUID, sa.ForeignKey("users.uuid"), nullable=False),
)


def verify_vfolder_name(folder: str) -> bool:
    if folder in RESERVED_VFOLDERS:
        return False
    for pattern in RESERVED_VFOLDER_PATTERNS:
        if pattern.match(folder):
            return False
    return True


async def query_accessible_vfolders(
    conn: SAConnection,
    user_uuid: uuid.UUID,
    *,
    # when enabled, skip vfolder ownership check if user role is admin or superadmin
    allow_privileged_access=False,
    user_role=None,
    domain_name=None,
    allowed_vfolder_types=None,
    extra_vf_conds=None,
    extra_invited_vf_conds=None,
    extra_vf_user_conds=None,
    extra_vf_group_conds=None,
) -> Sequence[Mapping[str, Any]]:
    from ai.backend.manager.models import association_groups_users as agus
    from ai.backend.manager.models import groups, users

    if allowed_vfolder_types is None:
        allowed_vfolder_types = ["user"]  # legacy default

    vfolders_selectors = [
        vfolders.c.name,
        vfolders.c.id,
        vfolders.c.host,
        vfolders.c.usage_mode,
        vfolders.c.created_at,
        vfolders.c.last_used,
        vfolders.c.max_files,
        vfolders.c.max_size,
        vfolders.c.ownership_type,
        vfolders.c.user,
        vfolders.c.group,
        vfolders.c.creator,
        vfolders.c.unmanaged_path,
        vfolders.c.cloneable,
        vfolders.c.status,
        # vfolders.c.permission,
        # users.c.email,
    ]

    async def _append_entries(_query, _is_owner=True):
        if extra_vf_conds is not None:
            _query = _query.where(extra_vf_conds)
        if extra_vf_user_conds is not None:
            _query = _query.where(extra_vf_user_conds)
        result = await conn.execute(_query)
        for row in result:
            row_keys = row.keys()
            _perm = (
                row.vfolder_permissions_permission
                if "vfolder_permissions_permission" in row_keys
                else row.vfolders_permission
            )
            entries.append(
                {
                    "name": row.vfolders_name,
                    "id": row.vfolders_id,
                    "host": row.vfolders_host,
                    "usage_mode": row.vfolders_usage_mode,
                    "created_at": row.vfolders_created_at,
                    "last_used": row.vfolders_last_used,
                    "max_size": row.vfolders_max_size,
                    "max_files": row.vfolders_max_files,
                    "ownership_type": row.vfolders_ownership_type,
                    "user": str(row.vfolders_user) if row.vfolders_user else None,
                    "group": str(row.vfolders_group) if row.vfolders_group else None,
                    "creator": row.vfolders_creator,
                    "user_email": row.users_email if "users_email" in row_keys else None,
                    "group_name": row.groups_name if "groups_name" in row_keys else None,
                    "is_owner": _is_owner,
                    "permission": _perm,
                    "unmanaged_path": row.vfolders_unmanaged_path,
                    "cloneable": row.vfolders_cloneable,
                    "status": row.vfolders_status,
                }
            )

    entries: List[dict] = []
    # User vfolders.
    if "user" in allowed_vfolder_types:
        # Scan my owned vfolders.
        j = vfolders.join(users, vfolders.c.user == users.c.uuid)
        query = sa.select(
            vfolders_selectors + [vfolders.c.permission, users.c.email], use_labels=True
        ).select_from(j)
        if not allow_privileged_access or (
            user_role != UserRole.ADMIN and user_role != UserRole.SUPERADMIN
        ):
            query = query.where(vfolders.c.user == user_uuid)
        await _append_entries(query)

        # Scan vfolders shared with me.
        j = vfolders.join(
            vfolder_permissions,
            vfolders.c.id == vfolder_permissions.c.vfolder,
            isouter=True,
        ).join(
            users,
            vfolders.c.user == users.c.uuid,
            isouter=True,
        )
        query = (
            sa.select(
                vfolders_selectors + [vfolder_permissions.c.permission, users.c.email],
                use_labels=True,
            )
            .select_from(j)
            .where(
                (vfolder_permissions.c.user == user_uuid)
                & (vfolders.c.ownership_type == VFolderOwnershipType.USER),
            )
        )
        if extra_invited_vf_conds is not None:
            query = query.where(extra_invited_vf_conds)
        await _append_entries(query, _is_owner=False)

    if "group" in allowed_vfolder_types:
        # Scan group vfolders.
        if user_role == UserRole.ADMIN or user_role == "admin":
            query = (
                sa.select([groups.c.id])
                .select_from(groups)
                .where(groups.c.domain_name == domain_name)
            )
            result = await conn.execute(query)
            grps = result.fetchall()
            group_ids = [g.id for g in grps]
        else:
            j = sa.join(agus, users, agus.c.user_id == users.c.uuid)
            query = sa.select([agus.c.group_id]).select_from(j).where(agus.c.user_id == user_uuid)
            result = await conn.execute(query)
            grps = result.fetchall()
            group_ids = [g.group_id for g in grps]
        j = vfolders.join(groups, vfolders.c.group == groups.c.id)
        query = sa.select(
            vfolders_selectors + [vfolders.c.permission, groups.c.name], use_labels=True
        ).select_from(j)
        if user_role != UserRole.SUPERADMIN:
            query = query.where(vfolders.c.group.in_(group_ids))
        if extra_vf_group_conds is not None:
            query = query.where(extra_vf_group_conds)
        is_owner = (user_role == UserRole.ADMIN or user_role == "admin") or (
            user_role == UserRole.SUPERADMIN or user_role == "superadmin"
        )
        await _append_entries(query, is_owner)

        # Override permissions, if exists, for group vfolders.
        j = sa.join(
            vfolders,
            vfolder_permissions,
            vfolders.c.id == vfolder_permissions.c.vfolder,
        )
        query = (
            sa.select(vfolder_permissions.c.permission, vfolder_permissions.c.vfolder)
            .select_from(j)
            .where(
                (vfolders.c.group.in_(group_ids)) & (vfolder_permissions.c.user == user_uuid),
            )
        )
        if extra_vf_conds is not None:
            query = query.where(extra_vf_conds)
        if extra_vf_user_conds is not None:
            query = query.where(extra_vf_user_conds)
        result = await conn.execute(query)
        overriding_permissions: dict = {row.vfolder: row.permission for row in result}
        for entry in entries:
            if (
                entry["id"] in overriding_permissions
                and entry["ownership_type"] == VFolderOwnershipType.GROUP
            ):
                entry["permission"] = overriding_permissions[entry["id"]]

    return entries


async def get_allowed_vfolder_hosts_by_group(
    conn: SAConnection,
    resource_policy,
    domain_name: str,
    group_id: Optional[uuid.UUID] = None,
    domain_admin: bool = False,
) -> VFolderHostPermissionMap:
    """
    Union `allowed_vfolder_hosts` from domain, group, and keypair_resource_policy.

    If `group_id` is not None, `allowed_vfolder_hosts` from the group is also merged.
    If the requester is a domain admin, gather all `allowed_vfolder_hosts` of the domain groups.
    """
    from . import domains, groups

    # Domain's allowed_vfolder_hosts.
    allowed_hosts = VFolderHostPermissionMap()
    query = sa.select([domains.c.allowed_vfolder_hosts]).where(
        (domains.c.name == domain_name) & (domains.c.is_active),
    )
    if values := await conn.scalar(query):
        allowed_hosts = allowed_hosts | values
    # Group's allowed_vfolder_hosts.
    if group_id is not None:
        query = sa.select([groups.c.allowed_vfolder_hosts]).where(
            (groups.c.domain_name == domain_name)
            & (groups.c.id == group_id)
            & (groups.c.is_active),
        )
        if values := await conn.scalar(query):
            allowed_hosts = allowed_hosts | values
    elif domain_admin:
        query = sa.select([groups.c.allowed_vfolder_hosts]).where(
            (groups.c.domain_name == domain_name) & (groups.c.is_active),
        )
        if rows := (await conn.execute(query)).fetchall():
            for row in rows:
                allowed_hosts = allowed_hosts | row.allowed_vfolder_hosts
    # Keypair Resource Policy's allowed_vfolder_hosts
    allowed_hosts = allowed_hosts | resource_policy["allowed_vfolder_hosts"]
    return allowed_hosts


async def get_allowed_vfolder_hosts_by_user(
    conn: SAConnection,
    resource_policy,
    domain_name: str,
    user_uuid: uuid.UUID,
    group_id: Optional[uuid.UUID] = None,
) -> VFolderHostPermissionMap:
    """
    Union `allowed_vfolder_hosts` from domain, groups, and keypair_resource_policy.

    All available `allowed_vfolder_hosts` of groups which requester associated will be merged.
    """
    from . import association_groups_users, domains, groups

    # Domain's allowed_vfolder_hosts.
    allowed_hosts = VFolderHostPermissionMap()
    query = sa.select([domains.c.allowed_vfolder_hosts]).where(
        (domains.c.name == domain_name) & (domains.c.is_active),
    )
    if values := await conn.scalar(query):
        allowed_hosts = allowed_hosts | values
    # User's Groups' allowed_vfolder_hosts.
    if group_id is not None:
        j = groups.join(
            association_groups_users,
            (
                (groups.c.id == association_groups_users.c.group_id)
                & (groups.c.id == group_id)
                & (association_groups_users.c.user_id == user_uuid)
            ),
        )
    else:
        j = groups.join(
            association_groups_users,
            (
                (groups.c.id == association_groups_users.c.group_id)
                & (association_groups_users.c.user_id == user_uuid)
            ),
        )
    query = (
        sa.select([groups.c.allowed_vfolder_hosts])
        .select_from(j)
        .where(
            (domains.c.name == domain_name) & (groups.c.is_active),
        )
    )
    if rows := (await conn.execute(query)).fetchall():
        for row in rows:
            allowed_hosts = allowed_hosts | row.allowed_vfolder_hosts
    # Keypair Resource Policy's allowed_vfolder_hosts
    allowed_hosts = allowed_hosts | resource_policy["allowed_vfolder_hosts"]
    return allowed_hosts


async def prepare_vfolder_mounts(
    conn: SAConnection,
    storage_manager: StorageSessionManager,
    allowed_vfolder_types: Sequence[str],
    user_scope: UserScope,
    requested_mounts: Sequence[str],
    requested_mount_map: Mapping[str, str],
) -> Sequence[VFolderMount]:
    """
    Determine the actual mount information from the requested vfolder lists,
    vfolder configurations, and the given user scope.
    """

    requested_vfolder_names: dict[str, str] = {}
    requested_vfolder_subpaths: dict[str, str] = {}
    requested_vfolder_dstpaths: dict[str, str] = {}
    matched_vfolder_mounts: list[VFolderMount] = []

    # Split the vfolder name and subpaths
    for key in requested_mounts:
        name, _, subpath = key.partition("/")
        if not PurePosixPath(os.path.normpath(key)).is_relative_to(name):
            raise InvalidAPIParameters(
                f"The subpath '{subpath}' should designate "
                f"a subdirectory of the vfolder '{name}'.",
            )
        requested_vfolder_names[key] = name
        requested_vfolder_subpaths[key] = os.path.normpath(subpath)
    for key, value in requested_mount_map.items():
        requested_vfolder_dstpaths[key] = value

    # Check if there are overlapping mount sources
    for p1 in requested_mounts:
        for p2 in requested_mounts:
            if p1 == p2:
                continue
            if PurePosixPath(p1).is_relative_to(PurePosixPath(p2)):
                raise InvalidAPIParameters(
                    f"VFolder source path '{p1}' overlaps with '{p2}'",
                )

    # Query the accessible vfolders that satisfy either:
    # - the name matches with the requested vfolder name, or
    # - the name starts with a dot (dot-prefixed vfolder) for automatic mounting.
    if requested_vfolder_names:
        extra_vf_conds = vfolders.c.name.in_(
            requested_vfolder_names.values()
        ) | vfolders.c.name.startswith(".")
    else:
        extra_vf_conds = vfolders.c.name.startswith(".")
    accessible_vfolders = await query_accessible_vfolders(
        conn,
        user_scope.user_uuid,
        user_role=user_scope.user_role,
        domain_name=user_scope.domain_name,
        allowed_vfolder_types=allowed_vfolder_types,
        extra_vf_conds=extra_vf_conds,
    )

    # Fast-path for empty requested mounts
    if not accessible_vfolders:
        if requested_vfolder_names:
            raise VFolderNotFound("There is no accessible vfolders at all.")
        else:
            return []
    accessible_vfolders_map = {vfolder["name"]: vfolder for vfolder in accessible_vfolders}

    # add automount folder list into requested_vfolder_names
    # and requested_vfolder_subpath
    for _vfolder in accessible_vfolders:
        if _vfolder["name"].startswith("."):
            requested_vfolder_names.setdefault(_vfolder["name"], _vfolder["name"])
            requested_vfolder_subpaths.setdefault(_vfolder["name"], ".")

    # for vfolder in accessible_vfolders:
    for key, vfolder_name in requested_vfolder_names.items():
        if not (vfolder := accessible_vfolders_map.get(vfolder_name)):
            raise VFolderNotFound(f"VFolder {vfolder_name} is not found or accessible.")
        if vfolder["group"] is not None and vfolder["group"] != str(user_scope.group_id):
            # User's accessible group vfolders should not be mounted
            # if not belong to the execution kernel.
            continue
        try:
            mount_base_path = PurePosixPath(
                await storage_manager.get_mount_path(
                    vfolder["host"],
                    vfolder["id"],
                    PurePosixPath(requested_vfolder_subpaths[key]),
                ),
            )
        except VFolderOperationFailed as e:
            raise InvalidAPIParameters(e.extra_msg, e.extra_data) from None
        if vfolder["name"] == ".local" and vfolder["group"] is not None:
            # Auto-create per-user subdirectory inside the group-owned ".local" vfolder.
            async with storage_manager.request(
                vfolder["host"],
                "POST",
                "folder/file/mkdir",
                params={
                    "volume": storage_manager.split_host(vfolder["host"])[1],
                    "vfid": vfolder["id"],
                    "relpath": str(user_scope.user_uuid.hex),
                    "exist_ok": True,
                },
            ):
                pass
            # Mount the per-user subdirectory as the ".local" vfolder.
            matched_vfolder_mounts.append(
                VFolderMount(
                    name=vfolder["name"],
                    vfid=vfolder["id"],
                    vfsubpath=PurePosixPath(user_scope.user_uuid.hex),
                    host_path=mount_base_path / user_scope.user_uuid.hex,
                    kernel_path=PurePosixPath("/home/work/.local"),
                    mount_perm=vfolder["permission"],
                )
            )
        else:
            # Normal vfolders
            kernel_path_raw = requested_vfolder_dstpaths.get(key)
            if kernel_path_raw is None:
                kernel_path = PurePosixPath(f"/home/work/{vfolder['name']}")
            else:
                kernel_path = PurePosixPath(kernel_path_raw)
                if not kernel_path.is_absolute():
                    kernel_path = PurePosixPath("/home/work", kernel_path_raw)
            matched_vfolder_mounts.append(
                VFolderMount(
                    name=vfolder["name"],
                    vfid=vfolder["id"],
                    vfsubpath=PurePosixPath(requested_vfolder_subpaths[key]),
                    host_path=mount_base_path / requested_vfolder_subpaths[key],
                    kernel_path=kernel_path,
                    mount_perm=vfolder["permission"],
                )
            )

    # Check if there are overlapping mount targets
    for vf1 in matched_vfolder_mounts:
        for vf2 in matched_vfolder_mounts:
            if vf1.name == vf2.name:
                continue
            if vf1.kernel_path.is_relative_to(vf2.kernel_path):
                raise InvalidAPIParameters(
                    f"VFolder mount path {vf1.kernel_path} overlaps with {vf2.kernel_path}",
                )

    return matched_vfolder_mounts


class VirtualFolder(graphene.ObjectType):
    class Meta:
        interfaces = (Item,)

    host = graphene.String()
    name = graphene.String()
    user = graphene.UUID()  # User.id (current owner, null in project vfolders)
    user_email = graphene.String()  # User.email (current owner, null in project vfolders)
    group = graphene.UUID()  # Group.id (current owner, null in user vfolders)
    group_name = graphene.String()  # Group.name (current owenr, null in user vfolders)
    creator = graphene.String()  # User.email (always set)
    unmanaged_path = graphene.String()
    usage_mode = graphene.String()
    permission = graphene.String()
    ownership_type = graphene.String()
    max_files = graphene.Int()
    max_size = BigInt()  # in MiB
    created_at = GQLDateTime()
    last_used = GQLDateTime()

    num_files = graphene.Int()
    cur_size = BigInt()
    # num_attached = graphene.Int()
    cloneable = graphene.Boolean()
    status = graphene.String()

    @classmethod
    def from_row(cls, ctx: GraphQueryContext, row: Row) -> Optional[VirtualFolder]:
        if row is None:
            return None
        return cls(
            id=row["id"],
            host=row["host"],
            name=row["name"],
            user=row["user"],
            user_email=row["users_email"],
            group=row["group"],
            group_name=row["groups_name"],
            creator=row["creator"],
            unmanaged_path=row["unmanaged_path"],
            usage_mode=row["usage_mode"],
            permission=row["permission"],
            ownership_type=row["ownership_type"],
            max_files=row["max_files"],
            max_size=row["max_size"],  # in MiB
            created_at=row["created_at"],
            last_used=row["last_used"],
            # num_attached=row['num_attached'],
            cloneable=row["cloneable"],
            status=row["status"],
        )

    async def resolve_num_files(self, info: graphene.ResolveInfo) -> int:
        # TODO: measure on-the-fly
        return 0

    async def resolve_cur_size(self, info: graphene.ResolveInfo) -> int:
        # TODO: measure on-the-fly
        return 0

    _queryfilter_fieldspec = {
        "id": ("vfolders_id", uuid.UUID),
        "host": ("vfolders_host", None),
        "name": ("vfolders_name", None),
        "group": ("vfolders_group", uuid.UUID),
        "group_name": ("groups_name", None),
        "user": ("vfolders_user", uuid.UUID),
        "user_email": ("users_email", None),
        "creator": ("vfolders_creator", None),
        "unmanaged_path": ("vfolders_unmanaged_path", None),
        "usage_mode": ("vfolders_usage_mode", lambda s: VFolderUsageMode[s]),
        "permission": ("vfolders_permission", lambda s: VFolderPermission[s]),
        "ownership_type": ("vfolders_ownership_type", lambda s: VFolderOwnershipType[s]),
        "max_files": ("vfolders_max_files", None),
        "max_size": ("vfolders_max_size", None),
        "created_at": ("vfolders_created_at", dtparse),
        "last_used": ("vfolders_last_used", dtparse),
        "cloneable": ("vfolders_cloneable", None),
        "status": ("vfolders_status", lambda s: VFolderOperationStatus[s]),
    }

    _queryorder_colmap = {
        "id": "vfolders_id",
        "host": "vfolders_host",
        "name": "vfolders_name",
        "group": "vfolders_group",
        "group_name": "groups_name",
        "user": "vfolders_user",
        "user_email": "users_email",
        "creator": "vfolders_creator",
        "usage_mode": "vfolders_usage_mode",
        "permission": "vfolders_permission",
        "ownership_type": "vfolders_ownership_type",
        "max_files": "vfolders_max_files",
        "max_size": "vfolders_max_size",
        "created_at": "vfolders_created_at",
        "last_used": "vfolders_last_used",
        "cloneable": "vfolders_cloneable",
        "status": "vfolders_status",
    }

    @classmethod
    async def load_count(
        cls,
        graph_ctx: GraphQueryContext,
        *,
        domain_name: str = None,
        group_id: uuid.UUID = None,
        user_id: uuid.UUID = None,
        filter: str = None,
    ) -> int:
        from .user import users

        j = vfolders.join(users, vfolders.c.user == users.c.uuid, isouter=True)
        query = sa.select([sa.func.count()]).select_from(j)
        if domain_name is not None:
            query = query.where(users.c.domain_name == domain_name)
        if group_id is not None:
            query = query.where(vfolders.c.group == group_id)
        if user_id is not None:
            query = query.where(vfolders.c.user == user_id)
        if filter is not None:
            qfparser = QueryFilterParser(cls._queryfilter_fieldspec)
            query = qfparser.append_filter(query, filter)
        async with graph_ctx.db.begin_readonly() as conn:
            result = await conn.execute(query)
            return result.scalar()

    @classmethod
    async def load_slice(
        cls,
        graph_ctx: GraphQueryContext,
        limit: int,
        offset: int,
        *,
        domain_name: str = None,
        group_id: uuid.UUID = None,
        user_id: uuid.UUID = None,
        filter: str = None,
        order: str = None,
    ) -> Sequence[VirtualFolder]:
        from .group import groups
        from .user import users

        j = vfolders.join(users, vfolders.c.user == users.c.uuid, isouter=True).join(
            groups, vfolders.c.group == groups.c.id, isouter=True
        )
        query = (
            sa.select([vfolders, users.c.email, groups.c.name.label("groups_name")])
            .select_from(j)
            .limit(limit)
            .offset(offset)
        )
        if domain_name is not None:
            query = query.where(users.c.domain_name == domain_name)
        if group_id is not None:
            query = query.where(vfolders.c.group == group_id)
        if user_id is not None:
            query = query.where(vfolders.c.user == user_id)
        if filter is not None:
            qfparser = QueryFilterParser(cls._queryfilter_fieldspec)
            query = qfparser.append_filter(query, filter)
        if order is not None:
            qoparser = QueryOrderParser(cls._queryorder_colmap)
            query = qoparser.append_ordering(query, order)
        else:
            query = query.order_by(vfolders.c.created_at.desc())
        async with graph_ctx.db.begin_readonly() as conn:
            return [
                obj
                async for r in (await conn.stream(query))
                if (obj := cls.from_row(graph_ctx, r)) is not None
            ]

    @classmethod
    async def batch_load_by_user(
        cls,
        graph_ctx: GraphQueryContext,
        user_uuids: Sequence[uuid.UUID],
        *,
        domain_name: str = None,
        group_id: uuid.UUID = None,
    ) -> Sequence[Sequence[VirtualFolder]]:
        from .user import users

        # TODO: num_attached count group-by
        j = sa.join(vfolders, users, vfolders.c.user == users.c.uuid)
        query = (
            sa.select([vfolders])
            .select_from(j)
            .where(vfolders.c.user.in_(user_uuids))
            .order_by(sa.desc(vfolders.c.created_at))
        )
        if domain_name is not None:
            query = query.where(users.c.domain_name == domain_name)
        if group_id is not None:
            query = query.where(vfolders.c.group == group_id)
        async with graph_ctx.db.begin_readonly() as conn:
            return await batch_multiresult(
                graph_ctx,
                conn,
                query,
                cls,
                user_uuids,
                lambda row: row["user"],
            )


class VirtualFolderList(graphene.ObjectType):
    class Meta:
        interfaces = (PaginatedList,)

    items = graphene.List(VirtualFolder, required=True)


class VirtualFolderPermission(graphene.ObjectType):
    class Meta:
        interfaces = (Item,)

    permission = graphene.String()
    vfolder = graphene.UUID()
    vfolder_name = graphene.String()
    user = graphene.UUID()
    user_email = graphene.String()

    @classmethod
    def from_row(cls, ctx: GraphQueryContext, row: Row) -> Optional[VirtualFolderPermission]:
        if row is None:
            return None
        return cls(
            permission=row["permission"],
            vfolder=row["vfolder"],
            vfolder_name=row["name"],
            user=row["user"],
            user_email=row["email"],
        )

    _queryfilter_fieldspec = {
        "permission": ("vfolder_permissions_permission", lambda s: VFolderPermission[s]),
        "vfolder": ("vfolder_permissions_vfolder", None),
        "vfolder_name": ("vfolders_name", None),
        "user": ("vfolder_permissions_user", None),
        "user_email": ("users_email", None),
    }

    _queryorder_colmap = {
        "permission": "vfolder_permissions_permission",
        "vfolder": "vfolder_permissions_vfolder",
        "vfolder_name": "vfolders_name",
        "user": "vfolder_permissions_user",
        "user_email": "users_email",
    }

    @classmethod
    async def load_count(
        cls,
        graph_ctx: GraphQueryContext,
        *,
        user_id: uuid.UUID = None,
        filter: str = None,
    ) -> int:
        from .user import users

        j = vfolder_permissions.join(vfolders, vfolders.c.id == vfolder_permissions.c.vfolder).join(
            users, users.c.uuid == vfolder_permissions.c.user
        )
        query = sa.select([sa.func.count()]).select_from(j)
        if user_id is not None:
            query = query.where(vfolders.c.user == user_id)
        if filter is not None:
            qfparser = QueryFilterParser(cls._queryfilter_fieldspec)
            query = qfparser.append_filter(query, filter)
        async with graph_ctx.db.begin_readonly() as conn:
            result = await conn.execute(query)
            return result.scalar()

    @classmethod
    async def load_slice(
        cls,
        graph_ctx: GraphQueryContext,
        limit: int,
        offset: int,
        *,
        user_id: uuid.UUID = None,
        filter: str = None,
        order: str = None,
    ) -> Sequence[VirtualFolderPermission]:
        from .user import users

        j = vfolder_permissions.join(vfolders, vfolders.c.id == vfolder_permissions.c.vfolder).join(
            users, users.c.uuid == vfolder_permissions.c.user
        )
        query = (
            sa.select([vfolder_permissions, vfolders.c.name, users.c.email])
            .select_from(j)
            .limit(limit)
            .offset(offset)
        )
        if user_id is not None:
            query = query.where(vfolders.c.user == user_id)
        if filter is not None:
            qfparser = QueryFilterParser(cls._queryfilter_fieldspec)
            query = qfparser.append_filter(query, filter)
        if order is not None:
            qoparser = QueryOrderParser(cls._queryorder_colmap)
            query = qoparser.append_ordering(query, order)
        else:
            query = query.order_by(vfolders.c.created_at.desc())
        async with graph_ctx.db.begin_readonly() as conn:
            return [
                obj
                async for r in (await conn.stream(query))
                if (obj := cls.from_row(graph_ctx, r)) is not None
            ]


class VirtualFolderPermissionList(graphene.ObjectType):
    class Meta:
        interfaces = (PaginatedList,)

    items = graphene.List(VirtualFolderPermission, required=True)



"""add-batch-session

Revision ID: c092dabf3ee5
Revises: c1409ad0e8da
Create Date: 2019-08-01 15:18:20.306290

"""
import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "c092dabf3ee5"
down_revision = "48ab2dfefba9"
branch_labels = None
depends_on = None


sessiontypes = postgresql.ENUM("INTERACTIVE", "BATCH", name="sessiontypes")


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("kernels", sa.Column("startup_command", sa.Text(), nullable=True))
    op.drop_column("kernels", "type")
    op.add_column(
        "kernels",
        sa.Column(
            "sess_type",
            sa.Enum("INTERACTIVE", "BATCH", name="sessiontypes"),
            nullable=False,
            server_default="INTERACTIVE",
        ),
    )
    op.create_index(op.f("ix_kernels_sess_type"), "kernels", ["sess_type"], unique=False)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix_kernels_sess_type"), table_name="kernels")
    op.drop_column("kernels", "sess_type")
    op.add_column(
        "kernels",
        sa.Column(
            "type",
            sa.Enum("INTERACTIVE", "BATCH", name="sessiontypes"),
            nullable=False,
            server_default="INTERACTIVE",
        ),
    )
    op.drop_column("kernels", "startup_command")
    # ### end Alembic commands ###



from __future__ import annotations

import logging
import uuid
from abc import ABCMeta, abstractmethod
from datetime import datetime
from typing import (
    Any,
    Dict,
    List,
    Mapping,
    MutableMapping,
    MutableSequence,
    Optional,
    Protocol,
    Sequence,
    Set,
)

import attrs
import sqlalchemy as sa
import trafaret as t
from sqlalchemy.engine.row import Row
from sqlalchemy.ext.asyncio import AsyncConnection as SAConnection
from sqlalchemy.sql import ColumnElement, Select

from ai.backend.common.docker import ImageRef
from ai.backend.common.logging import BraceStyleAdapter
from ai.backend.common.types import (
    AccessKey,
    AgentId,
    ClusterMode,
    KernelId,
    ResourceSlot,
    SessionId,
    SessionTypes,
    SlotName,
    SlotTypes,
    VFolderMount,
)

from ..defs import DEFAULT_ROLE
from ..models import kernels, keypairs
from ..models.scaling_group import ScalingGroupOpts
from ..registry import AgentRegistry

log = BraceStyleAdapter(logging.getLogger("ai.backend.manager.scheduler"))


def merge_resource(
    target: MutableMapping[str, Any],
    update: MutableMapping[str, Any],
) -> None:
    for k in update.keys():
        if k in target.keys():
            target[k] += update[k]
        else:
            target[k] = update[k]


@attrs.define(auto_attribs=True, slots=True)
class AgentAllocationContext:
    agent_id: Optional[AgentId]
    agent_addr: str
    scaling_group: str


@attrs.define(auto_attribs=True, slots=True)
class AgentContext:
    agent_id: AgentId
    agent_addr: str
    architecture: str
    scaling_group: str
    available_slots: ResourceSlot
    occupied_slots: ResourceSlot


@attrs.define(auto_attribs=True, slots=True)
class ScheduleDecision:
    agent_id: AgentId
    kernel_id: KernelId


@attrs.define(auto_attribs=True, slots=True)
class SchedulingContext:
    """
    Context for each scheduling decision.
    """

    registry: AgentRegistry
    known_slot_types: Mapping[SlotName, SlotTypes]


@attrs.define(auto_attribs=True, slots=True)
class ExistingSession:
    kernels: List[KernelInfo]
    access_key: AccessKey
    session_id: uuid.UUID
    session_type: SessionTypes
    session_name: str
    cluster_mode: ClusterMode
    cluster_size: int
    domain_name: str
    group_id: uuid.UUID
    scaling_group: str
    occupying_slots: ResourceSlot

    @classmethod
    def db_cols(cls) -> Set[ColumnElement]:
        return {
            kernels.c.id,
            kernels.c.status,
            kernels.c.access_key,
            kernels.c.session_id,
            kernels.c.session_type,
            kernels.c.session_name,
            kernels.c.cluster_mode,
            kernels.c.cluster_size,
            kernels.c.cluster_role,
            kernels.c.domain_name,
            kernels.c.group_id,
            kernels.c.scaling_group,
            kernels.c.occupied_slots,
        }

    @classmethod
    def base_query(cls) -> Select:
        return (
            sa.select(
                list(cls.db_cols() | KernelInfo.db_cols()),
            )
            .select_from(kernels)
            .order_by(kernels.c.created_at)
        )

    @classmethod
    def from_row(cls, row: Row) -> ExistingSession:
        return ExistingSession(
            kernels=[],
            access_key=row["access_key"],
            session_id=row["session_id"],
            session_type=row["session_type"],
            session_name=row["session_name"],
            cluster_mode=row["cluster_mode"],
            cluster_size=row["cluster_size"],
            domain_name=row["domain_name"],
            group_id=row["group_id"],
            scaling_group=row["scaling_group"],
            occupying_slots=ResourceSlot(),
        )

    @classmethod
    def from_rows(cls, rows: Sequence[Row]) -> List[ExistingSession]:
        items: Dict[str, ExistingSession] = {}
        for row in rows:
            if row["cluster_role"] == "main":
                items[row["session_id"]] = cls.from_row(row)
        for row in rows:
            session_id = row["session_id"]
            if session_id not in items:
                # In some cases, sub containers are still RUNNING
                # even though main container is TERMINATED.
                # To circumvent this edge case, we skip if main container
                # is not registered in `items`.
                continue
            session = items[session_id]
            session.kernels.append(KernelInfo.from_row(row))
            session.occupying_slots += row["occupied_slots"]  # type: ignore
        return list(items.values())


@attrs.define(auto_attribs=True, slots=True)
class PendingSession:
    """
    Context for individual session-related information used during scheduling.
    Resource parameters defined here should contain total amount of resources
    for all kernels in one session.
    """

    kernels: List[KernelInfo]
    access_key: AccessKey
    agent_id: AgentId
    agent_addr: str
    session_id: SessionId
    session_creation_id: str
    session_type: SessionTypes
    session_name: str
    cluster_mode: ClusterMode
    cluster_size: int
    domain_name: str
    group_id: uuid.UUID
    status_data: Mapping[str, Any]
    scaling_group: str
    resource_policy: str
    resource_opts: Mapping[str, Any]
    requested_slots: ResourceSlot
    target_sgroup_names: MutableSequence[str]
    environ: MutableMapping[str, str]
    vfolder_mounts: Sequence[VFolderMount]
    bootstrap_script: Optional[str]
    startup_command: Optional[str]
    internal_data: Optional[MutableMapping[str, Any]]
    preopen_ports: List[int]
    created_at: datetime
    use_host_network: bool

    @property
    def main_kernel_id(self) -> KernelId:
        for k in self.kernels:
            if k.cluster_role == DEFAULT_ROLE:
                return k.kernel_id
        raise RuntimeError("Unable to get the main kernel ID")

    @classmethod
    def db_cols(cls) -> Set[ColumnElement]:
        return {
            kernels.c.id,
            kernels.c.access_key,
            kernels.c.agent,
            kernels.c.agent_addr,
            kernels.c.session_creation_id,
            kernels.c.session_id,
            kernels.c.session_type,
            kernels.c.session_name,
            kernels.c.cluster_mode,
            kernels.c.cluster_size,
            kernels.c.domain_name,
            kernels.c.group_id,
            kernels.c.status_data,
            kernels.c.scaling_group,
            keypairs.c.resource_policy,
            kernels.c.occupied_slots,
            kernels.c.internal_data,
            kernels.c.resource_opts,
            kernels.c.environ,
            kernels.c.vfolder_mounts,
            kernels.c.bootstrap_script,
            kernels.c.startup_command,
            kernels.c.preopen_ports,
            kernels.c.created_at,
            kernels.c.use_host_network,
        }

    @classmethod
    def base_query(cls) -> Select:
        return (
            sa.select(
                list(cls.db_cols() | KernelInfo.db_cols()),
            )
            .select_from(
                sa.join(
                    kernels,
                    keypairs,
                    keypairs.c.access_key == kernels.c.access_key,
                )
            )
            .order_by(kernels.c.created_at)
        )

    @classmethod
    def from_row(cls, row: Row) -> PendingSession:
        return cls(
            kernels=[],
            access_key=row["access_key"],
            agent_id=row["agent"],
            agent_addr=row["agent_addr"],
            session_creation_id=row["session_creation_id"],
            session_id=row["session_id"],
            session_type=row["session_type"],
            session_name=row["session_name"],
            cluster_mode=row["cluster_mode"],
            cluster_size=row["cluster_size"],
            domain_name=row["domain_name"],
            group_id=row["group_id"],
            status_data=row["status_data"],
            scaling_group=row["scaling_group"],
            resource_policy=row["resource_policy"],
            resource_opts={},
            requested_slots=ResourceSlot(),
            internal_data=row["internal_data"],
            target_sgroup_names=[],
            environ={k: v for k, v in map(lambda s: s.split("=", maxsplit=1), row["environ"])},
            vfolder_mounts=row["vfolder_mounts"],
            bootstrap_script=row["bootstrap_script"],
            startup_command=row["startup_command"],
            preopen_ports=row["preopen_ports"],
            created_at=row["created_at"],
            use_host_network=row["use_host_network"],
        )

    @classmethod
    def from_rows(cls, rows: Sequence[Row]) -> List[PendingSession]:
        items: Dict[SessionId, PendingSession] = {}
        for row in rows:
            if row["cluster_role"] == "main":
                items[row["session_id"]] = cls.from_row(row)
        for row in rows:
            session = items[row["session_id"]]
            session.kernels.append(KernelInfo.from_row(row))
            session.requested_slots += row["occupied_slots"]  # type: ignore
            merge_resource(session.resource_opts, row["resource_opts"])  # type: ignore
        return list(items.values())


@attrs.define(auto_attribs=True, slots=True)
class KernelInfo:
    """
    Representing invididual kernel info.
    Resource parameters defined here should contain single value of resource
    for each kernel.
    """

    kernel_id: KernelId
    session_id: SessionId
    access_key: AccessKey
    agent_id: AgentId
    agent_addr: str
    cluster_role: str
    cluster_idx: int
    local_rank: int
    cluster_hostname: str
    image_ref: ImageRef
    resource_opts: Mapping[str, Any]
    requested_slots: ResourceSlot
    bootstrap_script: Optional[str]
    startup_command: Optional[str]
    created_at: datetime

    def __str__(self):
        return f"{self.kernel_id}#{self.cluster_role}{self.cluster_idx}"

    @classmethod
    def db_cols(cls) -> Set[ColumnElement]:
        return {
            kernels.c.id,
            kernels.c.session_id,
            kernels.c.access_key,
            kernels.c.agent,  # for scheduled kernels
            kernels.c.agent_addr,  # for scheduled kernels
            kernels.c.cluster_role,
            kernels.c.cluster_idx,
            kernels.c.local_rank,
            kernels.c.cluster_hostname,
            kernels.c.image,
            kernels.c.architecture,
            kernels.c.registry,
            kernels.c.resource_opts,
            kernels.c.occupied_slots,
            kernels.c.bootstrap_script,
            kernels.c.startup_command,
            kernels.c.created_at,
        }

    @classmethod
    def from_row(cls, row: Row) -> KernelInfo:
        return cls(
            kernel_id=row["id"],
            session_id=row["session_id"],
            access_key=row["access_key"],
            agent_id=row["agent"],
            agent_addr=row["agent_addr"],
            cluster_role=row["cluster_role"],
            cluster_idx=row["cluster_idx"],
            local_rank=row["local_rank"],
            cluster_hostname=row["cluster_hostname"],
            image_ref=ImageRef(row["image"], [row["registry"]], row["architecture"]),
            resource_opts=row["resource_opts"],
            requested_slots=row["occupied_slots"],
            bootstrap_script=row["bootstrap_script"],
            startup_command=row["startup_command"],
            created_at=row["created_at"],
        )


@attrs.define(auto_attribs=True, slots=True)
class KernelAgentBinding:
    kernel: KernelInfo
    agent_alloc_ctx: AgentAllocationContext
    allocated_host_ports: Set[int]


@attrs.define(auto_attribs=True, slots=True)
class PredicateResult:
    passed: bool
    message: Optional[str] = None


class SchedulingPredicate(Protocol):
    async def __call__(
        self,
        db_conn: SAConnection,
        sched_ctx: SchedulingContext,
        sess_ctx: PendingSession,
    ) -> PredicateResult:
        ...


class AbstractScheduler(metaclass=ABCMeta):

    """
    Interface for scheduling algorithms where the
    ``schedule()`` method is a pure function.
    """

    sgroup_opts: ScalingGroupOpts  # sgroup-specific config
    config: Mapping[str, Any]  # scheduler-specific config
    config_iv: t.Dict

    def __init__(self, sgroup_opts: ScalingGroupOpts, config: Mapping[str, Any]) -> None:
        self.sgroup_opts = sgroup_opts
        self.config = self.config_iv.check(config)

    @abstractmethod
    def pick_session(
        self,
        total_capacity: ResourceSlot,
        pending_sessions: Sequence[PendingSession],
        existing_sessions: Sequence[ExistingSession],
    ) -> Optional[SessionId]:
        """
        Pick a session to try schedule.
        This is where the queueing semantics is implemented such as prioritization.
        """
        return None

    @abstractmethod
    def assign_agent_for_session(
        self,
        possible_agents: Sequence[AgentContext],
        pending_session: PendingSession,
    ) -> Optional[AgentId]:
        """
        Assign an agent for the entire session, only considering the total requested
        slots of the session.  This is used for both single-container sessions and
        single-node multi-container sessions.

        In single-node multi-container sessions, all sub-containers are spawned by
        slicing the assigned agent's resource.
        """
        return None

    @abstractmethod
    def assign_agent_for_kernel(
        self,
        possible_agents: Sequence[AgentContext],
        pending_kernel: KernelInfo,
    ) -> Optional[AgentId]:
        """
        Assign an agent for a kernel of the session.
        This may be called multiple times for multi-node multi-container sessions.
        """
        return None



"""rename_mem_stats

Revision ID: e7371ca5797a
Revises: 93e9d31d40bf
Create Date: 2017-10-10 13:01:37.169568

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "e7371ca5797a"
down_revision = "93e9d31d40bf"
branch_labels = None
depends_on = None


def upgrade():
    op.alter_column("kernels", column_name="max_mem_bytes", new_column_name="mem_max_bytes")
    op.alter_column("kernels", column_name="cur_mem_bytes", new_column_name="mem_cur_bytes")


def downgrade():
    op.alter_column("kernels", column_name="mem_max_bytes", new_column_name="max_mem_bytes")
    op.alter_column("kernels", column_name="mem_cur_bytes", new_column_name="cur_mem_bytes")



"""add_vfolder_host_permission

Revision ID: 5bce905c21e5
Revises: 360af8f33d4e
Create Date: 2022-10-13 18:35:21.955941

"""
from typing import Dict, List

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql as pgsql
from sqlalchemy.sql.expression import bindparam

from ai.backend.manager.models.base import GUID, VFolderHostPermission, metadata

# revision identifiers, used by Alembic.
revision = "5bce905c21e5"
down_revision = "360af8f33d4e"
branch_labels = None
depends_on = None


ALL_HOST_PERMISSIONS = [perm.value for perm in VFolderHostPermission]


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    domains = sa.Table(
        "domains",
        metadata,
        sa.Column("name", sa.String(length=64), primary_key=True),
        sa.Column("allowed_vfolder_hosts", pgsql.ARRAY(sa.String)),
        extend_existing=True,
    )
    groups = sa.Table(
        "groups",
        metadata,
        sa.Column("id", GUID, primary_key=True),
        sa.Column("allowed_vfolder_hosts", pgsql.ARRAY(sa.String)),
        extend_existing=True,
    )
    keypair_resource_policies = sa.Table(
        "keypair_resource_policies",
        metadata,
        sa.Column("name", sa.String(length=256), primary_key=True),
        sa.Column("allowed_vfolder_hosts", pgsql.ARRAY(sa.String)),
        extend_existing=True,
    )
    domain_query = sa.select([domains.c.name, domains.c.allowed_vfolder_hosts])
    group_query = sa.select([groups.c.id, groups.c.allowed_vfolder_hosts])
    keypair_resource_policies_query = sa.select(
        [keypair_resource_policies.c.name, keypair_resource_policies.c.allowed_vfolder_hosts]
    )

    connection = op.get_bind()
    bind_param_pk = "bind_param_pk"

    def get_pk_vfolder_host_maps(query, pk_name) -> List:
        map_list = []
        rows = connection.execute(query).fetchall()
        for row in rows:
            hosts: List[str] = row["allowed_vfolder_hosts"]
            all_perms: Dict[str, List[str]] = {host: ALL_HOST_PERMISSIONS for host in hosts}
            map_list.append({bind_param_pk: row[pk_name], "allowed_vfolder_hosts": all_perms})
        return map_list

    domain_list = get_pk_vfolder_host_maps(domain_query, "name")
    group_list = get_pk_vfolder_host_maps(group_query, "id")
    resource_policy_list = get_pk_vfolder_host_maps(keypair_resource_policies_query, "name")

    op.drop_column("domains", "allowed_vfolder_hosts")
    op.drop_column("groups", "allowed_vfolder_hosts")
    op.drop_column("keypair_resource_policies", "allowed_vfolder_hosts")

    op.add_column(
        "domains",
        sa.Column(
            "allowed_vfolder_hosts",
            pgsql.JSONB(),
            nullable=False,
            server_default=sa.text("'{}'::jsonb"),
        ),
    )
    op.add_column(
        "groups",
        sa.Column(
            "allowed_vfolder_hosts",
            pgsql.JSONB(),
            nullable=False,
            server_default=sa.text("'{}'::jsonb"),
        ),
    )
    op.add_column(
        "keypair_resource_policies",
        sa.Column(
            "allowed_vfolder_hosts",
            pgsql.JSONB(),
            nullable=False,
            server_default=sa.text("'{}'::jsonb"),
        ),
    )

    domains = sa.Table(
        "domains",
        metadata,
        sa.Column("name", sa.String(length=64), primary_key=True),
        sa.Column("allowed_vfolder_hosts", pgsql.JSONB),
        extend_existing=True,
    )
    groups = sa.Table(
        "groups",
        metadata,
        sa.Column("id", GUID, primary_key=True),
        sa.Column("allowed_vfolder_hosts", pgsql.JSONB),
        extend_existing=True,
    )
    keypair_resource_policies = sa.Table(
        "keypair_resource_policies",
        metadata,
        sa.Column("name", sa.String(length=256), primary_key=True),
        sa.Column("allowed_vfolder_hosts", pgsql.JSONB),
        extend_existing=True,
    )

    def update_vfolder_host_col(table, cond, row_list):
        query = (
            sa.update(table)
            .values({"allowed_vfolder_hosts": bindparam("allowed_vfolder_hosts")})
            .where(cond)
        )
        connection.execute(query, row_list)

    update_vfolder_host_col(domains, domains.c.name == bindparam(bind_param_pk), domain_list)
    update_vfolder_host_col(groups, groups.c.id == bindparam(bind_param_pk), group_list)
    update_vfolder_host_col(
        keypair_resource_policies,
        keypair_resource_policies.c.name == bindparam(bind_param_pk),
        resource_policy_list,
    )

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    domains = sa.Table(
        "domains",
        metadata,
        sa.Column("name", sa.String(length=64), primary_key=True),
        sa.Column("allowed_vfolder_hosts", pgsql.JSONB),
        extend_existing=True,
    )
    groups = sa.Table(
        "groups",
        metadata,
        sa.Column("id", GUID, primary_key=True),
        sa.Column("allowed_vfolder_hosts", pgsql.JSONB),
        extend_existing=True,
    )
    keypair_resource_policies = sa.Table(
        "keypair_resource_policies",
        metadata,
        sa.Column("name", sa.String(length=256), primary_key=True),
        sa.Column("allowed_vfolder_hosts", pgsql.JSONB),
        extend_existing=True,
    )
    connection = op.get_bind()
    bind_param_pk = "bind_param_pk"

    def get_pk_vfolder_host_maps(query, pk_name) -> List:
        map_list = []
        rows = connection.execute(query).fetchall()
        for row in rows:
            hosts: Dict[str, List[str]] = row["allowed_vfolder_hosts"]
            map_list.append(
                {bind_param_pk: row[pk_name], "allowed_vfolder_hosts": list(hosts.keys())}
            )
        return map_list

    domain_query = sa.select([domains.c.name, domains.c.allowed_vfolder_hosts])
    group_query = sa.select([groups.c.id, groups.c.allowed_vfolder_hosts])
    keypair_resource_policies_query = sa.select(
        [keypair_resource_policies.c.name, keypair_resource_policies.c.allowed_vfolder_hosts]
    )

    domain_list = get_pk_vfolder_host_maps(domain_query, "name")
    group_list = get_pk_vfolder_host_maps(group_query, "id")
    resource_policy_list = get_pk_vfolder_host_maps(keypair_resource_policies_query, "name")

    op.drop_column("domains", "allowed_vfolder_hosts")
    op.drop_column("groups", "allowed_vfolder_hosts")
    op.drop_column("keypair_resource_policies", "allowed_vfolder_hosts")

    op.add_column(
        "domains",
        sa.Column(
            "allowed_vfolder_hosts", pgsql.ARRAY(sa.String), nullable=False, server_default="{}"
        ),
    )
    op.add_column(
        "groups",
        sa.Column(
            "allowed_vfolder_hosts", pgsql.ARRAY(sa.String), nullable=False, server_default="{}"
        ),
    )
    op.add_column(
        "keypair_resource_policies",
        sa.Column(
            "allowed_vfolder_hosts", pgsql.ARRAY(sa.String), nullable=False, server_default="{}"
        ),
    )

    domains = sa.Table(
        "domains",
        metadata,
        sa.Column("name", sa.String(length=64), primary_key=True),
        sa.Column("allowed_vfolder_hosts", pgsql.ARRAY(sa.String)),
        extend_existing=True,
    )
    groups = sa.Table(
        "groups",
        metadata,
        sa.Column("id", GUID, primary_key=True),
        sa.Column("allowed_vfolder_hosts", pgsql.ARRAY(sa.String)),
        extend_existing=True,
    )
    keypair_resource_policies = sa.Table(
        "keypair_resource_policies",
        metadata,
        sa.Column("name", sa.String(length=256), primary_key=True),
        sa.Column("allowed_vfolder_hosts", pgsql.ARRAY(sa.String)),
        extend_existing=True,
    )

    def update_vfolder_host_col(table, cond, row_list):
        query = (
            sa.update(table)
            .values({"allowed_vfolder_hosts": bindparam("allowed_vfolder_hosts")})
            .where(cond)
        )
        connection.execute(query, row_list)

    update_vfolder_host_col(domains, domains.c.name == bindparam(bind_param_pk), domain_list)
    update_vfolder_host_col(groups, groups.c.id == bindparam(bind_param_pk), group_list)
    update_vfolder_host_col(
        keypair_resource_policies,
        keypair_resource_policies.c.name == bindparam(bind_param_pk),
        resource_policy_list,
    )

    # ### end Alembic commands ###



"""add tag to kernels

Revision ID: d582942886ad
Revises: a1fd4e7b7782
Create Date: 2018-10-25 10:51:39.448309

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "d582942886ad"
down_revision = "a1fd4e7b7782"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("kernels", sa.Column("tag", sa.String(length=64), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("kernels", "tag")
    # ### end Alembic commands ###



"""add-scaling-group

Revision ID: 9a91532c8534
Revises: c401d78cc7b9
Create Date: 2019-07-25 22:32:25.974046

"""
import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql
from sqlalchemy.sql import text

import ai.backend.manager.models.base

# revision identifiers, used by Alembic.
revision = "9a91532c8534"
down_revision = "c401d78cc7b9"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "scaling_groups",
        sa.Column("name", sa.String(length=64), nullable=False),
        sa.Column("description", sa.String(length=512), nullable=True),
        sa.Column("is_active", sa.Boolean(), nullable=True),
        sa.Column(
            "created_at", sa.DateTime(timezone=True), server_default=sa.text("now()"), nullable=True
        ),
        sa.Column("driver", sa.String(length=64), nullable=False),
        sa.Column("driver_opts", postgresql.JSONB(astext_type=sa.Text()), nullable=False),
        sa.Column("scheduler", sa.String(length=64), nullable=False),
        sa.Column("scheduler_opts", postgresql.JSONB(astext_type=sa.Text()), nullable=False),
        sa.PrimaryKeyConstraint("name", name=op.f("pk_scaling_groups")),
    )
    op.create_index(
        op.f("ix_scaling_groups_is_active"), "scaling_groups", ["is_active"], unique=False
    )
    op.create_table(
        "sgroups_for_domains",
        sa.Column("scaling_group", sa.String(length=64), nullable=False),
        sa.Column("domain", sa.String(length=64), nullable=False),
        sa.ForeignKeyConstraint(
            ["domain"],
            ["domains.name"],
            name=op.f("fk_sgroups_for_domains_domain_domains"),
            onupdate="CASCADE",
            ondelete="CASCADE",
        ),
        sa.ForeignKeyConstraint(
            ["scaling_group"],
            ["scaling_groups.name"],
            name=op.f("fk_sgroups_for_domains_scaling_group_scaling_groups"),
            onupdate="CASCADE",
            ondelete="CASCADE",
        ),
        sa.UniqueConstraint("scaling_group", "domain", name="uq_sgroup_domain"),
    )
    op.create_index(
        op.f("ix_sgroups_for_domains_domain"), "sgroups_for_domains", ["domain"], unique=False
    )
    op.create_index(
        op.f("ix_sgroups_for_domains_scaling_group"),
        "sgroups_for_domains",
        ["scaling_group"],
        unique=False,
    )
    op.create_table(
        "sgroups_for_groups",
        sa.Column("scaling_group", sa.String(length=64), nullable=False),
        sa.Column("group", ai.backend.manager.models.base.GUID(), nullable=False),
        sa.ForeignKeyConstraint(
            ["group"],
            ["groups.id"],
            name=op.f("fk_sgroups_for_groups_group_groups"),
            onupdate="CASCADE",
            ondelete="CASCADE",
        ),
        sa.ForeignKeyConstraint(
            ["scaling_group"],
            ["scaling_groups.name"],
            name=op.f("fk_sgroups_for_groups_scaling_group_scaling_groups"),
            onupdate="CASCADE",
            ondelete="CASCADE",
        ),
        sa.UniqueConstraint("scaling_group", "group", name="uq_sgroup_ugroup"),
    )
    op.create_index(
        op.f("ix_sgroups_for_groups_group"), "sgroups_for_groups", ["group"], unique=False
    )
    op.create_index(
        op.f("ix_sgroups_for_groups_scaling_group"),
        "sgroups_for_groups",
        ["scaling_group"],
        unique=False,
    )
    op.create_table(
        "sgroups_for_keypairs",
        sa.Column("scaling_group", sa.String(length=64), nullable=False),
        sa.Column("access_key", sa.String(length=20), nullable=False),
        sa.ForeignKeyConstraint(
            ["access_key"],
            ["keypairs.access_key"],
            name=op.f("fk_sgroups_for_keypairs_access_key_keypairs"),
            onupdate="CASCADE",
            ondelete="CASCADE",
        ),
        sa.ForeignKeyConstraint(
            ["scaling_group"],
            ["scaling_groups.name"],
            name=op.f("fk_sgroups_for_keypairs_scaling_group_scaling_groups"),
            onupdate="CASCADE",
            ondelete="CASCADE",
        ),
        sa.UniqueConstraint("scaling_group", "access_key", name="uq_sgroup_akey"),
    )
    op.create_index(
        op.f("ix_sgroups_for_keypairs_access_key"),
        "sgroups_for_keypairs",
        ["access_key"],
        unique=False,
    )
    op.create_index(
        op.f("ix_sgroups_for_keypairs_scaling_group"),
        "sgroups_for_keypairs",
        ["scaling_group"],
        unique=False,
    )

    # create the default sgroup
    query = """
    INSERT INTO scaling_groups
    VALUES (
        'default',
        'The default agent scaling group',
        't',
        now(),
        'static',
        '{}'::jsonb,
        'fifo',
        '{}'::jsonb
    );
    """
    connection = op.get_bind()
    connection.execute(text(query))
    query = """
    INSERT INTO sgroups_for_domains
    VALUES ('default', 'default');
    """
    connection.execute(text(query))
    op.add_column(
        "agents",
        sa.Column("scaling_group", sa.String(length=64), server_default="default", nullable=False),
    )
    op.create_index(op.f("ix_agents_scaling_group"), "agents", ["scaling_group"], unique=False)
    op.create_foreign_key(
        op.f("fk_agents_scaling_group_scaling_groups"),
        "agents",
        "scaling_groups",
        ["scaling_group"],
        ["name"],
    )
    op.add_column(
        "kernels",
        sa.Column("scaling_group", sa.String(length=64), server_default="default", nullable=False),
    )
    op.create_index(op.f("ix_kernels_scaling_group"), "kernels", ["scaling_group"], unique=False)
    op.create_foreign_key(
        op.f("fk_kernels_scaling_group_scaling_groups"),
        "kernels",
        "scaling_groups",
        ["scaling_group"],
        ["name"],
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint(
        op.f("fk_kernels_scaling_group_scaling_groups"), "kernels", type_="foreignkey"
    )
    op.drop_index(op.f("ix_kernels_scaling_group"), table_name="kernels")
    op.drop_column("kernels", "scaling_group")
    op.drop_constraint(op.f("fk_agents_scaling_group_scaling_groups"), "agents", type_="foreignkey")
    op.drop_index(op.f("ix_agents_scaling_group"), table_name="agents")
    op.drop_column("agents", "scaling_group")
    op.drop_index(op.f("ix_sgroups_for_keypairs_scaling_group"), table_name="sgroups_for_keypairs")
    op.drop_index(op.f("ix_sgroups_for_keypairs_access_key"), table_name="sgroups_for_keypairs")
    op.drop_table("sgroups_for_keypairs")
    op.drop_index(op.f("ix_sgroups_for_groups_scaling_group"), table_name="sgroups_for_groups")
    op.drop_index(op.f("ix_sgroups_for_groups_group"), table_name="sgroups_for_groups")
    op.drop_table("sgroups_for_groups")
    op.drop_index(op.f("ix_sgroups_for_domains_scaling_group"), table_name="sgroups_for_domains")
    op.drop_index(op.f("ix_sgroups_for_domains_domain"), table_name="sgroups_for_domains")
    op.drop_table("sgroups_for_domains")
    op.drop_index(op.f("ix_scaling_groups_is_active"), table_name="scaling_groups")
    op.drop_table("scaling_groups")
    # ### end Alembic commands ###



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Dict, Sequence

import graphene
import sqlalchemy as sa
from sqlalchemy.engine.row import Row

from ai.backend.common.logging import BraceStyleAdapter
from ai.backend.common.types import ResourceSlot

from .base import (
    BigInt,
    BinarySize,
    ResourceSlotColumn,
    batch_result,
    metadata,
    set_if_set,
    simple_db_mutate,
    simple_db_mutate_returning_item,
)
from .user import UserRole

if TYPE_CHECKING:
    from .gql import GraphQueryContext

log = BraceStyleAdapter(logging.getLogger("ai.backend.manager.models"))

__all__: Sequence[str] = (
    "resource_presets",
    "ResourcePreset",
    "CreateResourcePreset",
    "ModifyResourcePreset",
    "DeleteResourcePreset",
)


resource_presets = sa.Table(
    "resource_presets",
    metadata,
    sa.Column("name", sa.String(length=256), primary_key=True),
    sa.Column("resource_slots", ResourceSlotColumn(), nullable=False),
    sa.Column("shared_memory", sa.BigInteger(), nullable=True),
)


class ResourcePreset(graphene.ObjectType):
    name = graphene.String()
    resource_slots = graphene.JSONString()
    shared_memory = BigInt()

    @classmethod
    def from_row(
        cls,
        ctx: GraphQueryContext,
        row: Row | None,
    ) -> ResourcePreset | None:
        if row is None:
            return None
        shared_memory = str(row["shared_memory"]) if row["shared_memory"] else None
        return cls(
            name=row["name"],
            resource_slots=row["resource_slots"].to_json(),
            shared_memory=shared_memory,
        )

    @classmethod
    async def load_all(cls, ctx: GraphQueryContext) -> Sequence[ResourcePreset]:
        query = sa.select([resource_presets]).select_from(resource_presets)
        async with ctx.db.begin_readonly() as conn:
            return [
                obj
                async for r in (await conn.stream(query))
                if (obj := cls.from_row(ctx, r)) is not None
            ]

    @classmethod
    async def batch_load_by_name(
        cls,
        ctx: GraphQueryContext,
        names: Sequence[str],
    ) -> Sequence[ResourcePreset | None]:
        query = (
            sa.select([resource_presets])
            .select_from(resource_presets)
            .where(resource_presets.c.name.in_(names))
            .order_by(resource_presets.c.name)
        )
        async with ctx.db.begin_readonly() as conn:
            return await batch_result(
                ctx,
                conn,
                query,
                cls,
                names,
                lambda row: row["name"],
            )


class CreateResourcePresetInput(graphene.InputObjectType):
    resource_slots = graphene.JSONString(required=True)
    shared_memory = graphene.String(required=False)


class ModifyResourcePresetInput(graphene.InputObjectType):
    resource_slots = graphene.JSONString(required=False)
    shared_memory = graphene.String(required=False)


class CreateResourcePreset(graphene.Mutation):

    allowed_roles = (UserRole.SUPERADMIN,)

    class Arguments:
        name = graphene.String(required=True)
        props = CreateResourcePresetInput(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()
    resource_preset = graphene.Field(lambda: ResourcePreset, required=False)

    @classmethod
    async def mutate(
        cls,
        root,
        info: graphene.ResolveInfo,
        name: str,
        props: CreateResourcePresetInput,
    ) -> CreateResourcePreset:
        data = {
            "name": name,
            "resource_slots": ResourceSlot.from_user_input(props.resource_slots, None),
            "shared_memory": BinarySize.from_str(props.shared_memory)
            if props.shared_memory
            else None,
        }
        insert_query = sa.insert(resource_presets).values(data)
        return await simple_db_mutate_returning_item(
            cls,
            info.context,
            insert_query,
            item_cls=ResourcePreset,
        )


class ModifyResourcePreset(graphene.Mutation):

    allowed_roles = (UserRole.SUPERADMIN,)

    class Arguments:
        name = graphene.String(required=True)
        props = ModifyResourcePresetInput(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()

    @classmethod
    async def mutate(
        cls,
        root,
        info: graphene.ResolveInfo,
        name: str,
        props: ModifyResourcePresetInput,
    ) -> ModifyResourcePreset:
        data: Dict[str, Any] = {}
        set_if_set(
            props,
            data,
            "resource_slots",
            clean_func=lambda v: ResourceSlot.from_user_input(v, None),
        )
        set_if_set(
            props, data, "shared_memory", clean_func=lambda v: BinarySize.from_str(v) if v else None
        )
        update_query = (
            sa.update(resource_presets).values(data).where(resource_presets.c.name == name)
        )
        return await simple_db_mutate(cls, info.context, update_query)


class DeleteResourcePreset(graphene.Mutation):

    allowed_roles = (UserRole.SUPERADMIN,)

    class Arguments:
        name = graphene.String(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()

    @classmethod
    async def mutate(
        cls,
        root,
        info: graphene.ResolveInfo,
        name: str,
    ) -> DeleteResourcePreset:
        delete_query = sa.delete(resource_presets).where(resource_presets.c.name == name)
        return await simple_db_mutate(cls, info.context, delete_query)
